from typing import Dict, Optional, List

from rest_framework.request import Request

from polaris.models import Asset, Transaction
from polaris.sep10.token import SEP10Token


class SEP31ReceiverIntegration:
    """
    The container class for SEP31 integrations
    """

    def info(
        self,
        request: Request,
        asset: Asset,
        lang: str = None,
        *args: List,
        **kwargs: Dict
    ) -> Dict:
        """
        .. _info response: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0031.md#response

        Return a dictionary containing the ``"fields"`` and `"sep12"` objects as described
        in the `info response`_ for the given `asset`. Polaris will provide the rest of the
        fields documented in the info response.

        Descriptions should be in the `lang` passed if supported.
        ::

            return {
                "fields": {
                    "transaction":{
                       "routing_number":{
                          "description": "routing number of the destination bank account"
                       },
                       "account_number":{
                          "description": "bank account number of the destination"
                       },
                       "type":{
                          "description": "type of deposit to make",
                          "choices":[
                             "SEPA",
                             "SWIFT"
                          ]
                       }
                    }
                },
                "sep12": {
                    "sender": {
                        "types": {
                            "sep31-sender": {
                                "description": "the only SEP-12 type for SEP-31 sending customers"
                            }
                        }
                    },
                    "receiver": {
                        "types": {
                            "sep31-receiver-cash-pickup": {
                                "description": "recipients who will pick up cash at physical locations"
                            },
                            "sep31-receiver-bank-transfer": {
                                "description" : "recipients who would like to receive funds via direct bank transfer"
                            }
                        }
                    }
                }
            }

        :param request: the ``rest_framework.request.Request`` instance
        :param asset: the ``Asset`` object for the field values returned
        :param lang: the ISO 639-1 language code of the user
        """
        raise NotImplementedError()

    def process_post_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args: List,
        **kwargs: Dict
    ) -> Optional[Dict]:
        """
        .. _customer-info-needed: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0031.md#customer-info-needed-400-bad-request
        .. _transaction-info-needed: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0031.md#transaction-info-needed-400-bad-request
        .. _SEP-12 GET /customer: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0012.md#customer-get

        Use the `params` passed in the request to do any processing of the user
        and requested transaction necessary to facilitate the payment to the
        receiving user. If the arguments are valid, save ``transaction`` and link it
        to your other models. If the transaction is saved but an error response is
        returned Polaris will return a 500 response to the user.

        If you'd like the user to send ``Transaction.amount_in`` `plus the fee amount`,
        add the amount charged as a fee to ``Transaction.amount_in`` and
        ``Transaction.amount_expected`` here.

        While not required per SEP-31, it is encouraged to also populate
        ``Transaction.amount_fee``, ``Transaction.fee_asset`` (if ``Transaction.quote``
        is not ``None``), and ``Transaction.amount_out`` here as well. Note that the
        amount sent over the Stellar Network could differ from the amount specified
        in this API call, so fees and the amount delievered may have to be recalculated
        in ``RailsIntegration.execute_outgoing_transaction()``.

        Also note that if your anchor service supports SEP-38, ``Transaction.quote``
        may be a firm or indicative ``Quote`` model instance representing the requested
        exchange of on and off-chain assets. If ``Quote.type == Quote.TYPE.indicative``,
        it is not yet saved to the database and Polaris will save it if a success response
        is returned.

        Polaris validates that the request includes all the required fields returned
        by ``SEP31ReceiverIntegration.info()`` but cannot validate the values. Return
        ``None`` if the params passed are valid, otherwise return one of the error
        dictionaries outlined below.

        If the `sender_id` or `receiver_id` values are invalid or the information
        collected for these users is not sufficient to process this request, return
        a dictionary matching the customer-info-needed_ response schema.
        ::

            return {
                "error": "customer_info_needed",
                "type": "sep31-large-amount-sender"
            }

        For example, the above response could be used if the anchor requires additional
        information on the sender when the `amount` is large. The `type` key specifies
        the appropriate type value the client should use for the sender's
        `SEP-12 GET /customer`_ request, and is optional.

        If some optional fields from ``info()`` are missing but needed for this
        transaction, return a dictionary matching the schema described in the
        transaction-info-needed_ response.
        ::

            return {
                "error": "transaction_info_needed",
                "fields": {
                    "transaction": {
                        "sender_bank_account": {
                            "description": (
                                "The bank account number used by the sender. "
                                "Only required for large transactions."
                            ),
                            "optional": True
                        }
                    }
                }
            }

        If some parameters passed are simply not acceptable, return a dictionary
        containing a single ``"error"`` key-value pair.
        ::

            return {
                "error": "invalid 'sender_bank_account' format"
            }

        :param token: The ``SEP10Token`` instance representing the authenticated session
        :param request: The ``rest_framework.request.Request`` instance
        :param params: The parameters included in the `/transaction` request
        :param transaction: the ``Transaction`` object representing the transaction being processed
        """
        raise NotImplementedError()

    def process_patch_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args: List,
        **kwargs: Dict
    ):
        """
        Use the `params` passed in the request to update `transaction` or any
        related data.

        Polaris validates that every field listed in
        ``Transaction.required_info_updates`` is present in `params` but
        cannot validate the values. If a ``ValueError`` is raised, Polaris will
        return a 400 response containing the exception message in the body.

        If no exception is raised, Polaris assumes the update was successful and
        will update the transaction's status back to ``pending_receiver`` as well
        as clear the ``required_info_updates`` and ``required_info_message`` fields.

        Once the transaction enters the ``pending_receiver`` status, the
        `execute_outgoing_transactions` process will attempt to send the payment
        to the receiving user. See the
        ``RailsIntegration.execute_outgoing_transaction`` function for more
        information on the lifecycle of a transaction.

        :param token: The ``SEP10Token`` instance representing the authenticated session
        :param request: the ``rest_framework.request.Request`` instance
        :param params: the request body of the `PATCH /transaction` request
        :param transaction: the ``Transaction`` object that should be updated
        """
        raise NotImplementedError()

    def valid_sending_anchor(
        self,
        token: SEP10Token,
        request: Request,
        public_key: str,
        *args: List,
        **kwargs: Dict
    ) -> bool:
        """
        Return ``True`` if `public_key` is a known anchor's stellar account address,
        and ``False`` otherwise. This function ensures that only registered sending
        anchors can make requests to protected endpoints.

        :param token: The ``SEP10Token`` instance representing the authenticated session
        :param request: the ``rest_framework.request.Request`` instance
        :param public_key: the public key of the sending anchor's stellar account
        """
        raise NotImplementedError()


registered_sep31_receiver_integration = SEP31ReceiverIntegration()



from typing import List, Dict
from django.db.models import QuerySet

from polaris.models import Transaction


class RailsIntegration:
    """
    A container class for functions that access off-chain rails, such banking
    accounts or other crypto networks.
    """

    def poll_outgoing_transactions(
        self, transactions: QuerySet, *args: List, **kwargs: Dict
    ) -> List[Transaction]:
        """
        Check the transactions that are still in a ``pending_external`` status and
        return the ones that are completed, meaning the user has received the funds.

        Polaris will update the transactions returned to ``Transaction.STATUS.completed``.

        `transactions` is passed as a Django ``QuerySet`` in case there are many pending
        transactions. You may want to query batches of ``Transaction`` objects to avoid
        consuming large amounts of memory.

        :param transactions: a ``QuerySet`` of ``Transaction`` objects
        """
        raise NotImplementedError()

    def execute_outgoing_transaction(
        self, transaction: Transaction, *args: List, **kwargs: Dict
    ):
        """
        Send the amount of the off-chain asset specified by `transaction` minus fees
        to the user associated with `transaction`. This function is used for SEP-6 &
        SEP-24 withdraws as well as SEP-31 payments. ``transaction.amount_fee`` and
        ``transaction.amount_out`` must be assigned in this function if they are not
        already. If the off-chain asset delivered to the user is different than the
        Stellar asset received on-chain, populate ``transaction.fee_asset`` as well.

        When this function is called, ``transaction.amount_in`` is the amount sent
        to the anchor, `not the amount specified in a SEP-24 or SEP-31 API call`.
        This matters because the amount actually sent to the anchor may differ from
        the amount specified in an API call. That is why you should always validate
        ``transaction.amount_in`` and recalculate ``transaction.amount_fee`` and
        ``transaction.amount_out`` here if necessary.

        If the amount is invalid in some way, the anchor must choose how to handle it.
        If you choose to refund the payment in its entirety, change ``transaction.status``
        to `"error"`, assign an appropriate message to ``transaction.status_message``,
        and update `transaction.refunded` to ``True``.

        You could also refund a portion of the amount and continue processing the
        remaining amount. In this case, the ``transaction.status`` column should be
        assigned one of the expected statuses for this function, mentioned below, and
        the ``amount_in`` field should be reassigned the value the anchor accepted.

        If the funds transferred to the user become available in the user's off-chain
        account immediately, or the anchor cannot verify when funds have become available,
        update ``Transaction.status`` to ``Transaction.STATUS.completed``. If the
        transfer was simply initiated and is pending external systems, update the status
        to ``Transaction.STATUS.pending_external``.

        If an exception is raised, the transaction will be left in
        its current status and may be used again as a parameter to this function.
        To ensure the exception isn't repeatedly re-raised, change the problematic
        transaction's status to ``Transaction.STATUS.error``.

        If ``transaction.protocol == Transaction.PROTOCOL.sep31`` and more information
        is required from the sending anchor or user to complete the transaction, update
        the status to ``Transaction.STATUS.pending_transaction_info_update`` and save a
        JSON-serialized string containing the fields that need updating to the
        ``Transaction.required_info_update`` column. The JSON string should be in the
        format returned from ``SEP31ReceiverIntegration.info()``. You can also
        optionally save a human-readable message to
        ``Transaction.required_info_message``. Both fields will be included in the
        `/transaction` response requested by the sending anchor.

        If the SEP-31 transaction is waiting for an update, the sending anchor will
        eventually make a request to the `PATCH /transaction` endpoint with the
        information specified in ``Transaction.required_info_update``. Once updated,
        this function will be called again with the updated transaction.

        :param transaction: the ``Transaction`` object associated with the payment
            this function should make
        """
        raise NotImplementedError()

    def poll_pending_deposits(
        self, pending_deposits: QuerySet, *args: List, **kwargs: Dict
    ) -> List[Transaction]:
        """
        .. _autocommit: https://docs.djangoproject.com/en/2.2/topics/db/transactions/#autocommit

        This function should poll the appropriate financial entity for the
        state of all `pending_deposits` and return the ones that have
        externally completed, meaning the off-chain funds are available in the
        anchor's account. For each returned transaction, ``Transaction.amount_fee``
        and ``Transaction.amount_out`` must be assigned. If the off-chain asset
        collected from the user is different than the Stellar asset to be sent on-chain,
        populate ``transaction.fee_asset`` as well.

        Client applications may send an amount that differs from the amount originally
        specified prior. If ``amount_expected`` differs from the amount deposited, the
        transaction should be placed in the ``error`` status or assign the amount
        deposited to ``amount_in`` and update ``amount_fee`` to appropriately.

        Also make sure to save the transaction's ``from_address`` field with
        the account the funds originated from.

        Any changes to the a transaction object must be saved to the database
        before it is returned. The transaction object will be refreshed using
        Django's ``.refresh_from_db()`` method and any unsaved data will be lost.

        For every transaction that is returned, Polaris will evaluate its readiness for
        submission to the Stellar network. If a transaction is completed on the network,
        the ``after_deposit()`` integration function will be called, however implementing
        this function is optional.

        `pending_deposits` is a QuerySet of the form
        ::

            pending_deposits = Transactions.object.filter(
                kind=Transaction.KIND.deposit,
                status=[
                    Transaction.STATUS.pending_user_transfer_start,
                    Transaction.STATUS.pending_external
                ],
                pending_execution_attempt=False
            )

        ``pending_user_transfer_start`` is the proper status for a
        transaction when the user must take some action to proceed. In this
        case, that action is sending the deposit funds.

        ``pending_external`` is the proper status for a transaction when
        the deposit funds have been sent but have not arrived in the anchor's
        off-chain account. If the anchor cannot detect when deposit funds
        are sent but not received, it is perfectly acceptable to keep the
        transaction in ``pending_user_transfer_start`` until the funds have
        arrived.

        **Note**: As of verison 1.3, this function is called within a database
        transaction context manager:
        ::

            with django.db.transaction.atomic():
                ready_transactions = rri.poll_pending_deposits(
                    pending_deposits.select_for_update()
                )
                Transaction.objects.filter(
                    id__in=[t.id for t in ready_transactions]
                ).update(pending_execution_attempt=True)

        This is done to ensure the same ``Transaction`` object is not retrieved
        from the database by multiple invocations of the process_pending_deposits
        command and submitted to Stellar as unique transactions.

        This differs from the majority of other queries Polaris makes, which
        are executed in autocommit_ mode, the Django default.

        :param pending_deposits: a django Queryset for pending Transactions
        :return: a list of ``Transaction`` objects which correspond to
            successful user deposits to the anchor's account.
        """
        raise NotImplementedError()


registered_rails_integration = RailsIntegration()



from typing import List, Dict

from rest_framework.request import Request

from polaris.models import Asset


def get_stellar_toml(request: Request, *args: List, **kwargs: Dict):
    """
    .. _SEP-1: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0001.md
    .. _`Account Info`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0001.md#account-information

    Replace this function with another by passing it to ``register_integrations()``.

    The dictionary returned will be merged with Polaris' default attributes and serialized
    using the ``toml.dumps()`` function. The output will be rendered in the HTTP response.

    The base attributes provided by Polaris are:

    - `ACCOUNTS`
    - `VERSION`
    - `SIGNING_KEY`
    - `NETWORK_PASSPHRASE`
    - `WEB_AUTH_ENDPOINT`
    - `TRANSFER_SERVER`
    - `TRANSFER_SERVER_0024`
    - `KYC_SERVER`
    - `DIRECT_PAYMENT_SERVER`
    - `QUOTE_SERVER`

    The contents of the dictionary returned will overwrite the default matching key values.

    :return: a dictionary of SEP-1_ attributes
    """
    return {
        "CURRENCIES": [
            {"code": asset.code, "issuer": asset.issuer}
            for asset in Asset.objects.all()
        ]
    }


registered_toml_func = get_stellar_toml



from typing import Dict, Optional
from decimal import Decimal, InvalidOperation
from collections import defaultdict

from django.utils.translation import (
    gettext as _,
    get_supported_language_variant,
)
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from rest_framework.views import APIView
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer
from rest_framework.parsers import JSONParser

from polaris.locale.utils import activate_lang_for_request
from polaris.sep10.utils import validate_sep10_token
from polaris.sep10.token import SEP10Token
from polaris.integrations import (
    registered_sep31_receiver_integration,
    registered_custody_integration as rci,
)
from polaris.models import Transaction, Asset, Quote
from polaris.sep31.serializers import SEP31TransactionSerializer
from polaris.utils import (
    getLogger,
    get_quote_and_offchain_destination_asset,
    render_error_response,
    create_transaction_id,
    validate_patch_request_fields,
    validate_account_and_memo,
)


logger = getLogger(__name__)


class TransactionsAPIView(APIView):
    parser_classes = [JSONParser]
    renderer_classes = [JSONRenderer, BrowsableAPIRenderer]

    @staticmethod
    @validate_sep10_token()
    def get(
        token: SEP10Token,
        request: Request,
        transaction_id: str = None,
    ) -> Response:
        if not transaction_id:
            return render_error_response(
                _("GET requests must include a transaction ID in the URI"),
            )
        elif not registered_sep31_receiver_integration.valid_sending_anchor(
            token=token, request=request, public_key=token.account
        ):
            return render_error_response(_("invalid sending account."), status_code=403)
        elif not transaction_id:
            return render_error_response(_("missing 'id' in URI"))
        try:
            t = (
                Transaction.objects.filter(
                    id=transaction_id,
                    stellar_account=token.account,
                    protocol=Transaction.PROTOCOL.sep31,
                )
                .select_related("quote")
                .first()
            )
        except ValidationError:  # bad id parameter
            return render_error_response(_("transaction not found"), status_code=404)
        if not t:
            return render_error_response(_("transaction not found"), status_code=404)
        return Response({"transaction": SEP31TransactionSerializer(t).data})

    @staticmethod
    @validate_sep10_token()
    def patch(
        token: SEP10Token,
        request: Request,
        transaction_id: str = None,
    ) -> Response:
        if not registered_sep31_receiver_integration.valid_sending_anchor(
            token=token, request=request, public_key=token.account
        ):
            return render_error_response(_("invalid sending account"), status_code=403)
        try:
            transaction = Transaction.objects.filter(
                id=transaction_id,
                stellar_account=token.account,
                protocol=Transaction.PROTOCOL.sep31,
            ).get()
        except (ValidationError, ObjectDoesNotExist):
            return render_error_response(_("transaction not found"), status_code=404)
        if transaction.status != Transaction.STATUS.pending_transaction_info_update:
            return render_error_response(_("update not required"))
        try:
            validate_patch_request_fields(request.data.get("fields"), transaction)
            registered_sep31_receiver_integration.process_patch_request(
                token=token,
                request=request,
                params=request.data.get("fields"),
                transaction=transaction,
            )
        except ValueError as e:
            return render_error_response(str(e))
        except RuntimeError as e:
            logger.exception(str(e))
            return render_error_response(
                _("unable to process request"), status_code=500
            )
        transaction.status = Transaction.STATUS.pending_receiver
        transaction.required_info_updates = None
        transaction.required_info_message = None
        transaction.save()
        return Response(status=200)

    @staticmethod
    @validate_sep10_token()
    def post(
        token: SEP10Token,
        request: Request,
        **kwargs: Dict,
    ):
        if kwargs:
            return render_error_response(
                _("POST requests should not specify subresources in the URI")
            )
        elif not registered_sep31_receiver_integration.valid_sending_anchor(
            token=token, request=request, public_key=token.account
        ):
            return render_error_response("invalid sending account", status_code=403)

        try:
            params = validate_post_request(token, request)
        except ValueError as e:
            return render_error_response(str(e))

        # validate fields separately since error responses need different format
        missing_fields = validate_post_fields(
            request, params.get("fields"), params.get("asset"), params.get("lang")
        )
        if missing_fields:
            return Response(
                {"error": "transaction_info_needed", "fields": missing_fields},
                status=400,
            )

        # create transaction object without saving to the DB
        transaction = Transaction(
            id=create_transaction_id(),
            protocol=Transaction.PROTOCOL.sep31,
            kind=Transaction.KIND.send,
            status=Transaction.STATUS.pending_sender,
            stellar_account=token.account,
            asset=params["asset"],
            amount_in=params["amount"],
            amount_expected=params["amount"],
            client_domain=token.client_domain,
            quote=params["quote"],
        )

        error_data = registered_sep31_receiver_integration.process_post_request(
            token=token, request=request, params=params, transaction=transaction
        )
        try:
            validate_error_response(error_data, transaction)
        except ValueError as e:
            logger.error(str(e))
            return render_error_response(
                _("unable to process the request"), status_code=500
            )
        if error_data:
            return Response(error_data, status=400)

        if transaction.quote and transaction.quote.type == Quote.TYPE.indicative:
            transaction.quote.save()
        try:
            receiving_account, memo, memo_type = validate_account_and_memo(
                *rci.get_receiving_account_and_memo(
                    request=request, transaction=transaction
                )
            )
        except ValueError:
            logger.exception(
                "CustodyIntegration.get_receiving_account_and_memo() returned invalid values"
            )
            return render_error_response(
                _("unable to process the request"), status_code=500
            )
        transaction.receiving_anchor_account = receiving_account
        transaction.memo = memo
        transaction.memo_type = memo_type
        transaction.save()
        return Response(
            {
                "id": transaction.id,
                "stellar_account_id": transaction.receiving_anchor_account,
                "stellar_memo": transaction.memo,
                "stellar_memo_type": transaction.memo_type,
            },
            status=201,
        )


def validate_post_request(token: SEP10Token, request: Request) -> Dict:
    asset_args = {"code": request.data.get("asset_code")}
    if request.data.get("asset_issuer"):
        asset_args["issuer"] = request.data.get("asset_issuer")
    asset = Asset.objects.filter(**asset_args).first()
    if not (asset and asset.sep31_enabled):
        raise ValueError(_("invalid 'asset_code' and 'asset_issuer'"))
    try:
        amount = round(Decimal(request.data.get("amount")), asset.significant_decimals)
    except (InvalidOperation, TypeError):
        raise ValueError(_("invalid 'amount'"))
    if asset.send_min_amount > amount or amount > asset.send_max_amount:
        raise ValueError(_("invalid 'amount'"))
    lang = request.data.get("lang")
    if lang:
        try:
            get_supported_language_variant(lang)
        except LookupError:
            raise ValueError("unsupported 'lang'")
        activate_lang_for_request(lang)
    if not isinstance(request.data.get("fields"), dict):
        raise ValueError(_("'fields' must serialize to a JSON object"))
    elif len(request.data["fields"]) not in [0, 1]:
        raise ValueError("'fields' should only have one key, 'transaction'")
    elif request.data["fields"] and not isinstance(
        request.data["fields"].get("transaction"), dict
    ):
        raise ValueError(_("'transaction' value in 'fields' must be a dict"))
    if not (
        type(request.data.get("sender_id")) in [str, type(None)]
        and type(request.data.get("receiver_id")) in [str, type(None)]
    ):
        raise ValueError(_("'sender_id' and 'receiver_id' values must be strings"))
    quote, destination_asset = get_quote_and_offchain_destination_asset(
        token=token,
        quote_id=request.data.get("quote_id"),
        destination_asset_str=request.data.get("destination_asset"),
        asset=asset,
        amount=Decimal(amount),
    )
    if (
        quote
        and quote.type == Quote.TYPE.firm
        and Transaction.objects.filter(quote=quote).exists()
    ):
        raise ValueError(_("quote has already been used in a transaction"))
    return {
        "asset": asset,
        "amount": amount,
        "lang": lang,
        # fields are validated in validate_fields()
        "sender_id": request.data.get("sender_id"),
        "receiver_id": request.data.get("receiver_id"),
        "fields": request.data.get("fields"),
        "destination_asset": destination_asset,
        "quote": quote,
    }


def validate_post_fields(
    request: Request, passed_fields: Dict, asset: Asset, lang: Optional[str]
) -> Dict:
    missing_fields = defaultdict(dict)
    expected_fields = registered_sep31_receiver_integration.info(
        request=request, asset=asset, lang=lang
    ).get("fields", {})
    if "transaction" not in expected_fields:
        return {}
    for field, info in expected_fields["transaction"].items():
        if info.get("optional"):
            continue
        elif field not in passed_fields.get("transaction", {}):
            missing_fields["transaction"][field] = info
    return dict(missing_fields)


def validate_error_response(error_data: Dict, transaction: Transaction):
    if not error_data:
        return
    if Transaction.objects.filter(id=transaction.id).exists():
        raise ValueError(f"transactions should not be created on bad requests")
    elif error_data["error"] == "transaction_info_needed":
        if not isinstance(error_data.get("fields"), dict):
            raise ValueError("'fields' must serialize to a JSON object")
        validate_post_fields_needed(error_data.get("fields"), transaction.asset)
        if len(error_data) > 2:
            raise ValueError("extra fields returned in customer_info_needed response")
    elif error_data["error"] == "customer_info_needed":
        if ("type" in error_data and len(error_data) > 2) or (
            "type" not in error_data and len(error_data) > 1
        ):
            raise ValueError(
                "extra fields returned in transaction_info_needed response"
            )
        elif type(error_data.get("type")) not in [None, str]:
            raise ValueError("invalid value for 'type' key")
    elif not isinstance(error_data["error"], str):
        raise ValueError("'error' must be a string")


def validate_post_fields_needed(response_fields: Dict, asset: Asset):
    expected_fields = registered_sep31_receiver_integration.info(asset).get(
        "fields", {}
    )
    if "transaction" not in response_fields:
        raise ValueError("unrecognized category of fields in response")
    if not isinstance(response_fields["transaction"], dict):
        raise ValueError("'transaction' value must be a dict")
    for field, value in response_fields["transaction"].items():
        if field not in expected_fields["transaction"]:
            raise ValueError(f"unrecognized field in 'transaction' object")
        elif not (
            isinstance(response_fields["transaction"][field], dict)
            and response_fields["transaction"][field].get("description")
        ):
            raise ValueError(f"field value must be a dict with a description")



# Generated by Django 3.2.7 on 2021-09-18 20:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0011_transaction_amount_expected"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="account_memo",
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="transaction",
            name="muxed_account",
            field=models.TextField(blank=True, null=True),
        ),
    ]



from jwt import InvalidIssuerError
import urllib3
from decimal import Decimal
from typing import Optional, Dict
from logging import getLogger

from django.core.management.base import BaseCommand, CommandError
from django.db.models import Q
from stellar_sdk import Keypair, TransactionBuilder, Server
from stellar_sdk import Asset as StellarSdkAsset
from stellar_sdk.account import Account
from stellar_sdk.exceptions import (
    NotFoundError,
    BaseHorizonError,
    Ed25519SecretSeedInvalidError,
)

from polaris import settings
from polaris.models import Transaction, Asset

logger = getLogger(__name__)


class Command(BaseCommand):
    """
    The testnet command comes with the following commands:

    ``issue`` allows users to create assets on the Stellar testnet network. When
    the test network resets, you’ll have to reissue your assets.

    ``reset`` calls the functionality invoked by ``issue`` for each asset in the
    anchor’s database. Since the database does not store the issuing account’s
    secret key, the user must input each key as requested by the Polaris command.
    It also performs a couple other functions necessary to ensure your Polaris
    instance runs successfully after a testnet reset:

    Moves all pending_trust transactions to error
        This is done because all accounts have been cleared from the network. While
        its possible an account that required a trustline could be recreated and a
        trustline could be established, its unlikely. Polaris assumes a testnet
        reset makes in-progress transactions unrecoverable.

    Updates the paging_token of latest transaction streamed for each anchored asset
        :mod:`~polaris.management.commands.watch_transactions` streams transactions to
        and from each anchored asset’s distribution account. Specifically, it streams
        transactions starting with the most recently completed transaction’s
        :attr:`~polaris.models.Transaction.paging_token` on startup. When the testnet
        resets, the :attr:`~polaris.models.Transaction.paging_token` used for transactions
        prior to the reset are no longer valid. To fix this, Polaris updates the
        :attr:`~polaris.models.Transaction.paging_token` of the most recently completed
        transaction for each anchored asset to "now".

    **Positional arguments:**

    reset
        -h, --help  show this help message and exit

    issue
        -h, --help          show this help message and exit
        --asset ASSET, -a ASSET
                            the code of the asset issued
        --issuer-seed ISSUER_SEED, -i ISSUER_SEED
                            the issuer's secret key
        --distribution-seed DISTRIBUTION_SEED, -d DISTRIBUTION_SEED
                            the distribution account's secret key
        --client-seed CLIENT_SEED, -c CLIENT_SEED
                            the client account's secret key
        --issue-amount ISSUE_AMOUNT
                            the amount sent to distribution account. Also the
                            limit for the trustline.
        --client-amount CLIENT_AMOUNT
                            the amount sent to client account. Also the limit for
                            the trustline.

    ``add-asset`` allows users to add assets to the database:

    add-asset
        -h, --help           show this help message and exit
        --asset ASSET, -a ASSET
                             the code of the asset issued
        --issuer-public ISSUER_PUBLIC, -i ISSUER_PUBLIC
                             the issuer's public key
        --distribution-seed  DISTRIBUTION_SEED, -d DISTRIBUTION_SEED
                             the distribution account's secret key
        --sep6-enabled       (Optional) flag to enable sep6 for this asset, defaults to false
        --sep24-enabled      (Optional) flag to enable sep24 for this asset, defaults to false
        --sep31-enabled      (Optional) flag to enable sep31 for this asset, defaults to false
        --sep38-enabled      (Optional) flag to enable sep38 for this asset, defaults to false
        --deposit-enabled    (Optional) flag to enable deposits for this asset, defaults to false
        --withdrawal-enabled (Optional) flag to enable withdrawals for this asset, defaults to false
        --symbol             (Optional) symbol for the asset, default to "$"

    ``delete-asset`` allows users to delete assets from the database:

    delete-asset
        -h, --help            show this help message and exit
        --asset ASSET, -a ASSET
                              the code of the asset to be deleted
        --issuer-public ISSUER_PUBLIC, -i ISSUER_PUBLIC
                              the issuer's public key
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset_parser = None
        self.issue_parser = None
        self.add_asset_parser = None
        self.delete_asset_parser = None
        self.server = Server("https://horizon-testnet.stellar.org")
        self.http = urllib3.PoolManager()

    def add_arguments(self, parser):  # pragma: no cover
        subparsers = parser.add_subparsers(dest="subcommands", required=True)
        self.reset_parser = subparsers.add_parser(
            "reset", help="a sub-command for testnet resets"
        )
        self.issue_parser = subparsers.add_parser(
            "issue", help="a sub-command for issuing assets on testnet"
        )
        self.issue_parser.add_argument(
            "--asset", "-a", help="the code of the asset issued"
        )
        self.issue_parser.add_argument(
            "--issuer-seed", "-i", help="the issuer's secret key"
        )
        self.issue_parser.add_argument(
            "--distribution-seed", "-d", help="the distribution account's secret key"
        )
        self.issue_parser.add_argument(
            "--client-seed", "-c", help="the client account's secret key"
        )
        self.issue_parser.add_argument(
            "--issue-amount",
            type=Decimal,
            help="the amount sent to distribution account. Also the limit for the trustline.",
        )
        self.issue_parser.add_argument(
            "--client-amount",
            type=Decimal,
            help="the amount sent to client account. Also the limit for the trustline.",
        )

        self.add_asset_parser = subparsers.add_parser(
            "add-asset", help="a sub-command for adding assets to the database"
        )
        self.add_asset_parser.add_argument(
            "--asset", help="asset code to add to the database", required=True
        )
        self.add_asset_parser.add_argument(
            "--issuer-public", help="the asset issuer's public key", required=True
        )
        self.add_asset_parser.add_argument(
            "--distribution-seed",
            help="the anchors distribution seed for this asset",
            required=True,
        )
        self.add_asset_parser.add_argument(
            "--sep6-enabled",
            help="enable sep6 for this asset",
            action="store_true",
        )
        self.add_asset_parser.add_argument(
            "--sep24-enabled",
            help="enable sep24 for this asset",
            action="store_true",
        )
        self.add_asset_parser.add_argument(
            "--sep31-enabled",
            help="enable sep31 for this asset",
            action="store_true",
        )
        self.add_asset_parser.add_argument(
            "--sep38-enabled",
            help="enable sep38 for this asset",
            action="store_true",
        )
        self.add_asset_parser.add_argument(
            "--deposit-enabled",
            help="enable deposits for this asset",
            action="store_true",
        )
        self.add_asset_parser.add_argument(
            "--withdrawal-enabled",
            help="enable withdrawals for this asset",
            action="store_true",
        )
        self.add_asset_parser.add_argument(
            "--symbol",
            help="symbol for the asset",
        )

        self.delete_asset_parser = subparsers.add_parser(
            "delete-asset", help="a sub-command for deleting assets to the database"
        )
        self.delete_asset_parser.add_argument(
            "--asset", help="asset code to delete to the database", required=True
        )
        self.delete_asset_parser.add_argument(
            "--issuer-public", help="the asset issuer's public key", required=True
        )

    def handle(self, *_args, **options):  # pragma: no cover
        if options.get("subcommands") == "reset":
            self.reset(**options)
        elif options.get("subcommands") == "issue":
            self.issue(**options)
        elif options.get("subcommands") == "add-asset":
            self.add_asset_to_db(**options)
        elif options.get("subcommands") == "delete-asset":
            self.delete_asset_from_db(**options)

    def reset(self, **options):
        """
        Perform any necessary functions to ensure the anchor is in a valid
        state after a testnet reset. Currently this involves the following:

        - re-issuing every Asset object in the DB
        - moves all `pending_trust` Transactions to `error`
        - setting the most-recently streamed Transaction object's
            `paging_token` attribute to None. This signals to
            watch_transactions to stream using the `"now"` keyword.
        """
        print("\nResetting each asset's most recent paging token")
        for asset in Asset.objects.filter(distribution_seed__isnull=False):
            transaction = (
                Transaction.objects.filter(
                    Q(kind=Transaction.KIND.withdrawal) | Q(kind=Transaction.KIND.send),
                    receiving_anchor_account=asset.distribution_account,
                    status=Transaction.STATUS.completed,
                )
                .order_by("-completed_at")
                .first()
            )
            if transaction:
                transaction.paging_token = None
                transaction.save()
        print("\nPlacing all pending_trust transactions into error status")
        Transaction.objects.filter(status=Transaction.STATUS.pending_trust).update(
            status=Transaction.STATUS.error
        )
        for asset in Asset.objects.filter(issuer__isnull=False):
            print(f"\nIssuing {asset.code}")
            issuer_seed = input(f"Seed for {asset.code} issuer (enter to skip): ")
            if not issuer_seed:
                continue
            try:
                Keypair.from_secret(issuer_seed)
            except Ed25519SecretSeedInvalidError:
                raise CommandError("Bad seed string for issuer account")
            distribution_seed = asset.distribution_seed
            if not distribution_seed:
                distribution_seed = input(
                    f"Seed for {asset.code} distribution account: "
                )
                try:
                    Keypair.from_secret(distribution_seed)
                except Ed25519SecretSeedInvalidError:
                    raise CommandError("Bad seed string for distribution account")
            self.issue(
                **{
                    "asset": asset.code,
                    "issuer_seed": issuer_seed,
                    "distribution_seed": distribution_seed,
                    "issue_amount": Decimal(10000000),
                }
            )

    def get_or_create_accounts(self, account_keypairs: list):
        """
        Get account details from horizon for the specified account_keypairs,
        if the account doesnt exist, create the account
        """
        accounts = {}
        for kp in account_keypairs:
            if not kp:
                continue
            try:
                accounts[kp.public_key] = (
                    self.server.accounts().account_id(kp.public_key).call()
                )
            except NotFoundError:
                print(f"Funding {kp.public_key} ...")
                self.http.request(
                    "GET", f"https://friendbot.stellar.org?addr={kp.public_key}"
                )
                accounts[kp.public_key] = (
                    self.server.accounts().account_id(kp.public_key).call()
                )
        return accounts

    def issue(self, **options):
        """
        Issue the asset specified using the `options` passed to the subcommand
        or function call. The code here is a port of the following project:

        https://github.com/stellar/create-stellar-token

        Users can setup distribution and client accounts for the asset being
        issued as well.
        """
        code = options.get("asset") or "TEST"
        issuer = Keypair.from_secret(
            options.get("issuer_seed") or Keypair.random().secret
        )
        distributor = Keypair.from_secret(
            options.get("distribution_seed") or Keypair.random().secret
        )
        client, client_amount = None, None
        if options.get("client_seed") or options.get("client_amount"):
            client = Keypair.from_secret(
                options.get("client_seed") or Keypair.random().secret
            )
            client_amount = options.get("client_amount") or Decimal(1000)
        issue_amount = options.get("issue_amount") or Decimal(100000)

        print("\nIssuer account public and private keys:")
        print(f"public key: {issuer.public_key}")
        print(f"secret key: {issuer.secret}\n")
        print("Distribution account public and private keys:")
        print(f"public key: {distributor.public_key}")
        print(f"secret key: {distributor.secret}\n")
        print(f"Issuing {issue_amount} {code} to the distribution account")
        if client:
            print("\nClient account public and private keys:")
            print(f"public key: {client.public_key}")
            print(f"secret key: {client.secret}\n")
            print(f"Sending {client_amount} to the client account\n")

        accounts = self.get_or_create_accounts([issuer, distributor, client])

        self.add_balance(code, issue_amount, accounts, distributor, issuer, issuer)
        if client:
            self.add_balance(code, client_amount, accounts, client, distributor, issuer)

        home_domain = input("Home domain for the issuing account (enter to skip): ")
        if home_domain:
            self.set_home_domain(issuer, home_domain)

    def set_home_domain(self, issuer: Keypair, home_domain: str):
        envelope = (
            TransactionBuilder(
                self.server.load_account(issuer.public_key),
                base_fee=settings.MAX_TRANSACTION_FEE_STROOPS
                or self.server.fetch_base_fee(),
                network_passphrase="Test SDF Network ; September 2015",
            )
            .append_set_options_op(home_domain=home_domain)
            .set_timeout(30)
            .build()
        )
        envelope.sign(issuer)
        try:
            self.server.submit_transaction(envelope)
        except BaseHorizonError as e:
            print(
                f"Failed to set {home_domain} as home_domain for {issuer.public_key}."
                f"Result codes: {e.extras.get('result_codes')}"
            )
        else:
            print("Success!")

    def add_balance(self, code, amount, accounts, dest, src, issuer):
        tb = TransactionBuilder(
            self.server.load_account(src.public_key),
            base_fee=settings.MAX_TRANSACTION_FEE_STROOPS
            or self.server.fetch_base_fee(),
            network_passphrase="Test SDF Network ; September 2015",
        )
        balance = self.get_balance(code, issuer.public_key, accounts[dest.public_key])
        if not balance:
            print(f"\nCreating {code} trustline for {dest.public_key}")
            if settings.MAX_TRANSACTION_FEE_STROOPS:
                # halve the base_fee because there are 2 operations
                tb.base_fee = tb.base_fee // 2
            tb.append_change_trust_op(
                asset=StellarSdkAsset(code=code, issuer=issuer.public_key),
                source=dest.public_key,
            )
            payment_amount = amount
        elif Decimal(balance) < amount:
            print(f"\nReplenishing {code} balance to {amount} for {dest.public_key}")
            payment_amount = amount - Decimal(balance)
        else:
            print(
                "Destination account already has more than the amount "
                "specified, skipping\n"
            )
            return

        print(f"Sending {code} payment of {payment_amount} to {dest.public_key}")
        tb.append_payment_op(
            destination=dest.public_key,
            amount=payment_amount,
            asset=StellarSdkAsset(code=code, issuer=issuer.public_key),
            source=src.public_key,
        )
        envelope = tb.set_timeout(30).build()
        if len(tb.operations) == 2:
            # add destination's signature if we're adding a trustline
            envelope.sign(dest)
        envelope.sign(src)
        try:
            self.server.submit_transaction(envelope)
        except BaseHorizonError as e:
            print(
                f"Failed to send {code} payment to {dest.public_key}. "
                f"Result codes: {e.extras.get('result_codes')}"
            )
        else:
            print("Success!")

    def get_balance(self, code, issuer_public_key, json) -> Optional[str]:
        for balance_obj in json["balances"]:
            if (
                balance_obj.get("asset_code") == code
                and balance_obj.get("asset_issuer") == issuer_public_key
            ):
                return balance_obj["balance"]

    def add_asset_to_db(self, **options):
        """
        Add an asset to the database
        """
        asset_code = options.get("asset")
        issuer = options.get("issuer_public")
        distribution_seed = options.get("distribution_seed")
        if not (issuer and asset_code):
            print(f"\ninvalid value in options {options}")
            return
        for asset in Asset.objects.filter(issuer__isnull=False):
            if asset.code == asset_code:
                print(
                    f"\nAsset {asset.code}:{asset.issuer} already exists in the database, skipping..."
                )
                return

        print(f"\nAdding asset {asset_code}:{issuer} to database")
        Asset.objects.create(
            code=asset_code,
            issuer=issuer,
            distribution_seed=distribution_seed,
            sep6_enabled=options.get("sep6_enabled") or False,
            sep24_enabled=options.get("sep24_enabled") or False,
            sep31_enabled=options.get("sep31_enabled") or False,
            sep38_enabled=options.get("sep38_enabled") or False,
            deposit_enabled=options.get("deposit_enabled") or False,
            withdrawal_enabled=options.get("withdrawal_enabled") or False,
            symbol=options.get("symbol") or "$",
        )
        print(f"\nAsset {asset_code}:{issuer} added to database!")

    def delete_asset_from_db(self, **options):
        """
        Delete an asset from the database
        """
        asset_code = options.get("asset")
        issuer = options.get("issuer_public")

        asset_to_delete = Asset.objects.filter(code=asset_code, issuer=issuer).first()
        if not asset_to_delete:
            print(f"\nNo asset matching {asset_code}:{issuer} found to delete")
            return

        asset_to_delete.delete()
        print(f"\nAsset: {asset_code}:{issuer} deleted from the database")



import time
from datetime import datetime

import jwt
import pytz
from jwt import InvalidTokenError, ExpiredSignatureError
from urllib.parse import urlencode
from typing import Callable, Dict, Optional
from decimal import Decimal, DecimalException

from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response
from django.core.exceptions import (
    ValidationError,
    ImproperlyConfigured,
    ObjectDoesNotExist,
)
from django.utils.translation import gettext as _
from django.urls import reverse
from django.conf import settings as django_settings
from django.core.validators import URLValidator
from stellar_sdk import MuxedAccount

from polaris.locale.utils import (
    activate_lang_for_request,
    validate_or_use_default_language,
)
from polaris import settings
from polaris.utils import getLogger
from polaris.models import Asset, Transaction
from polaris.utils import render_error_response, verify_valid_asset_operation


logger = getLogger(__name__)


def check_authentication(as_html: bool = True) -> Callable:
    """
    Authentication decorator for POST /interactive endoints
    """

    def decorator(view) -> Callable:
        def wrapper(request: Request, *_args, **_kwargs) -> Response:
            try:
                check_authentication_helper(request)
            except ValueError as e:
                return render_error_response(str(e), as_html=as_html, status_code=403)
            else:
                return view(request)

        return wrapper

    return decorator


def authenticate_session(as_html: bool = True) -> Callable:
    """
    Authentication decorator for GET /interactive endpoints
    """

    def decorator(view) -> Callable:
        def wrapper(request: Request, *_args, **_kwargs) -> Response:
            try:
                authenticate_session_helper(request)
            except ValueError as e:
                return render_error_response(str(e), as_html=as_html, status_code=403)
            else:
                return view(request)

        return wrapper

    return decorator


def authenticate_session_helper(r: Request):
    """
    Decodes and validates the JWT token passed in the GET request to a
    /webapp endpoint.

    Adds two items to ``r.session``:
    - authenticated: a boolean for whether the session has been authenticated
        for any transaction
    - account: the stellar account address associated with the token
    """
    # Don't authenticate in local mode, since session cookies will not be
    # included in request/response headers without HTTPS
    if settings.LOCAL_MODE:
        return

    token = r.GET.get("token")
    if not token:
        # If there is no token, check if this session has already been authenticated,
        # that the session's account is the one that initiated the transaction, and
        # that the session has been authenticated for this particular transaction.
        if r.session.get("authenticated") and r.session.get("account"):
            tid = r.GET.get("transaction_id")
            tqs = Transaction.objects.filter(
                id=tid,
                stellar_account=r.session["account"],
                muxed_account=r.session["muxed_account"],
                account_memo=r.session["memo"],
            )
            if not (tid in r.session.get("transactions", []) and tqs.exists()):
                raise ValueError(f"Not authenticated for transaction ID: {tid}")
            else:  # pragma: no cover
                # client has been authenticated for the requested transaction
                return
        else:
            raise ValueError("Missing authentication token")
    elif r.session.exists(r.session.session_key) and token in r.session.get(
        "tokens", []
    ):
        # The client opened an interactive flow with a specific token for the second time
        raise ValueError("Unexpected one-time auth token")
    elif r.session.get("tokens"):
        r.session["tokens"].append(token)
    else:
        r.session["tokens"] = [token]

    try:
        jwt_dict = jwt.decode(token, settings.SERVER_JWT_KEY, algorithms=["HS256"])
    except ExpiredSignatureError:
        raise ValueError(_("Your session has expired. Please restart the transaction"))
    except InvalidTokenError as e:
        raise ValueError(str(e))

    now = time.time()
    if jwt_dict["iss"] != r.build_absolute_uri("interactive"):
        raise ValueError(_("Invalid token issuer"))
    elif jwt_dict["iat"] > now or jwt_dict["exp"] < now:
        raise ValueError(_("Token is not yet valid or is expired"))

    if jwt_dict["sub"].startswith("M"):
        muxed_account = jwt_dict["sub"]
        stellar_account = MuxedAccount.from_account(muxed_account).account_id
        account_memo = None
    elif ":" in jwt_dict["sub"]:
        stellar_account, account_memo = jwt_dict["sub"].split(":")
        muxed_account = None
    else:
        stellar_account = jwt_dict["sub"]
        account_memo = None
        muxed_account = None

    transaction_qs = Transaction.objects.filter(
        id=jwt_dict["jti"],
        stellar_account=stellar_account,
        muxed_account=muxed_account,
        account_memo=account_memo,
    )
    if not transaction_qs.exists():
        raise ValueError(_("Transaction for account not found"))

    # JWT is valid, authenticate session
    r.session["authenticated"] = True
    r.session["account"] = stellar_account
    r.session["muxed_account"] = muxed_account
    r.session["memo"] = account_memo
    try:
        r.session["transactions"].append(jwt_dict["jti"])
    except KeyError:
        r.session["transactions"] = [jwt_dict["jti"]]

    # persists the session, generating r.session.session_key
    #
    # this session key is passed to the rendered views and
    # used in client-side JavaScript in requests to the server
    if not r.session.session_key:
        r.session.create()


def check_authentication_helper(r: Request):
    """
    Checks that the session associated with the request is authenticated
    """
    # Don't authenticate in local mode, since session cookies will not be
    # included in request/response headers without HTTPS
    if settings.LOCAL_MODE:
        return

    if not r.session.get("authenticated"):
        raise ValueError(_("Session is not authenticated"))

    try:
        if not Transaction.objects.filter(
            id=r.GET.get("transaction_id"), stellar_account=r.session.get("account")
        ).exists():
            raise ValueError(_("Transaction for account not found"))
    except ValidationError:
        raise ValueError(_("Transaction for account not found"))


def invalidate_session(request: Request):
    """
    Invalidates request's session for the interactive flow.
    """
    # Don't try to invalidate a session in local mode, since session cookies
    # were never included in request/response headers
    if settings.LOCAL_MODE:
        return

    request.session["authenticated"] = False


def validate_url(url) -> Optional[Dict]:
    schemes = ["https"] if not settings.LOCAL_MODE else ["https", "http"]
    try:
        URLValidator(schemes=schemes)(url)
    except ValidationError:
        return dict(
            error=render_error_response(
                _("invalid callback URL provided"), as_html=True
            )
        )


def interactive_args_validation(request: Request, kind: str) -> Dict:
    """
    Validates the arguments passed to the /webapp endpoints

    Returns a dictionary, either containing an 'error' response
    object or the transaction and asset objects specified by the
    incoming request.
    """
    transaction_id = request.GET.get("transaction_id")
    asset_code = request.GET.get("asset_code")
    callback = request.GET.get("callback")
    on_change_callback = request.GET.get("on_change_callback")
    amount_str = request.GET.get("amount")
    lang = validate_or_use_default_language(request.GET.get("lang"))
    activate_lang_for_request(lang)
    asset = Asset.objects.filter(code=asset_code, sep24_enabled=True).first()
    if not transaction_id:
        return dict(
            error=render_error_response(_("no 'transaction_id' provided"), as_html=True)
        )
    elif not (asset_code and asset):
        return dict(
            error=render_error_response(_("invalid 'asset_code'"), as_html=True)
        )
    elif on_change_callback and any(
        domain in on_change_callback
        for domain in settings.CALLBACK_REQUEST_DOMAIN_DENYLIST
    ):
        on_change_callback = None
    try:
        transaction = Transaction.objects.get(id=transaction_id, asset=asset, kind=kind)
    except (ObjectDoesNotExist, ValidationError):
        return dict(
            error=render_error_response(
                _("Transaction with ID and asset_code not found"),
                as_html=True,
                status_code=status.HTTP_404_NOT_FOUND,
            )
        )

    # Verify that amount is provided, and that can be parsed into a decimal:
    amount = None
    if amount_str:
        try:
            amount = Decimal(amount_str)
        except (DecimalException, TypeError):
            return dict(
                error=render_error_response(_("invalid 'amount'"), as_html=True)
            )

        err_resp = verify_valid_asset_operation(
            asset, amount, transaction.kind, as_html=True
        )
        if err_resp:
            return dict(error=err_resp)

    for url in [callback, on_change_callback]:
        if url and url.lower() != "postmessage":
            error_response = validate_url(url)
            if error_response:
                return error_response

    return dict(
        transaction=transaction,
        asset=asset,
        callback=callback,
        on_change_callback=on_change_callback,
        amount=amount,
        lang=lang,
    )


def generate_interactive_jwt(
    request: Request, transaction_id: str, account: str, memo: int
) -> str:
    """
    Generates a 30-second JWT for the client to use in the GET URL for
    the interactive flow.
    """
    issued_at = int(time.time()) - 1
    payload = {
        "iss": request.build_absolute_uri(request.path),
        "iat": issued_at,
        "exp": issued_at + settings.INTERACTIVE_JWT_EXPIRATION,
        "sub": f"{account}:{memo}" if memo else account,
        "jti": transaction_id,
    }
    return jwt.encode(payload, settings.SERVER_JWT_KEY, algorithm="HS256")


def check_sep24_config():
    check_middleware()
    check_protocol()


def check_middleware():
    """
    Ensures the Django app running Polaris has the correct middleware
    configuration. Polaris requires SessionMiddleware to be installed.
    """
    session_middleware_path = "django.contrib.sessions.middleware.SessionMiddleware"
    if session_middleware_path not in django_settings.MIDDLEWARE:
        raise ImproperlyConfigured(
            f"{session_middleware_path} is not installed in settings.MIDDLEWARE"
        )
    if not settings.LOCAL_MODE and not getattr(
        django_settings, "SESSION_COOKIE_SECURE", False
    ):
        raise ImproperlyConfigured(
            "the SESSION_COOKIE_SECURE setting must be set to True"
        )


def check_protocol():
    if settings.LOCAL_MODE:
        logger.warning(
            "Polaris is in local mode. This makes the SEP-24 interactive flow "
            "insecure and should only be used for local development."
        )
        if getattr(django_settings, "SECURE_SSL_REDIRECT"):
            logger.warning(
                "Using SECURE_SSL_REDIRECT while in local mode does not make "
                "interactive flows secure."
            )


def interactive_url(
    request: Request,
    transaction_id: str,
    account: str,
    memo: int,
    asset_code: str,
    op_type: str,
    amount: Optional[Decimal],
    lang: Optional[str],
) -> Optional[str]:
    params = {
        "asset_code": asset_code,
        "transaction_id": transaction_id,
        "token": generate_interactive_jwt(request, transaction_id, account, memo),
    }
    if amount:
        params["amount"] = amount
    if lang:
        params["lang"] = lang
    qparams = urlencode(params)
    if op_type == settings.OPERATION_WITHDRAWAL:
        url_params = f"{reverse('get_interactive_withdraw')}?{qparams}"
    else:
        url_params = f"{reverse('get_interactive_deposit')}?{qparams}"
    return request.build_absolute_uri(url_params)


def get_timezone_utc_offset(timezone) -> int:
    return round(
        datetime.now().astimezone(pytz.timezone(timezone)).utcoffset().total_seconds()
        / 60
    )



from django import forms
from django.utils.formats import localize
from django.utils.translation import gettext_lazy as _
from django.forms.widgets import TextInput

from polaris.models import Transaction, Asset


class CardNumberInput(TextInput):
    template_name = "widgets/card_number.html"


class CardExpirationInput(TextInput):
    template_name = "widgets/card_expiration.html"


class CardCvvInput(TextInput):
    template_name = "widgets/card_cvv.html"


class CreditCardField(forms.CharField):
    def __init__(self, placeholder=None, *args, **kwargs):
        super().__init__(
            # override default widget
            widget=CardNumberInput(attrs={"placeholder": placeholder}),
            *args,
            **kwargs,
        )

    default_error_messages = {
        "invalid": _("The credit card number is invalid"),
    }

    @staticmethod
    def luhn_checksum(card_number):
        def digits_of(n):
            return [int(d) for d in str(n)]

        digits = digits_of(card_number)
        odd_digits = digits[-1::-2]
        even_digits = digits[-2::-2]
        checksum = 0
        checksum += sum(odd_digits)
        for ed in even_digits:
            checksum += sum(digits_of(ed * 2))
        return checksum % 10

    def is_luhn_valid(self, card_number):
        return self.luhn_checksum(card_number) == 0

    def clean(self, value):
        # ensure no spaces or dashes
        value = value.replace(" ", "").replace("-", "")
        if not (value.isdigit() and self.is_luhn_valid(value)):
            raise forms.ValidationError(self.error_messages["invalid"])
        return value


class CreditCardForm(forms.Form):
    """
    A generic form for collecting credit or debit card information.

    Ensures `card_number` is valid, but does not validate the `expiration` or
    `cvv`. Subclass this form for additional validation.
    """

    name = forms.CharField(label=_("Name"))
    card_number = CreditCardField(label=_("Card Number"))
    expiration = forms.Field(widget=CardExpirationInput, label=_("Expiration"))
    cvv = forms.Field(widget=CardCvvInput, label=_("CVV"))


class TransactionForm(forms.Form):
    """
    A base class for collecting transaction information. Developers must define
    subclasses to collect additional information and apply additional validation.

    This form assumes the amount collected is in units of a Stellar
    :class:`~polaris.models.Asset`. If the amount of an
    :class:`~polaris.models.OffChainAsset` must be collected, create a different
    form.

    Note that Polaris' base UI treats the amount field on this form and its
    subclasses differently than other forms. Specifically, Polaris automatically
    adds the asset's symbol to the input field, adds a placeholder value of 0,
    makes the fee table visible (by default), and uses the amount entered to update
    the fee table on each change.

    If you do not want the fee table to be displayed when this form class is
    rendered, set ``"show_fee_table"`` to ``False`` in the dict returned from
    :meth:`~polaris.integrations.DepositIntegration.content_for_template`.

    Fee calculation within the UI is done using the asset's fixed and percentage
    fee values saved to the database. If those values are not present, Polaris makes
    calls to the anchor's `/fee` endpoint and displays the response value to the
    user. If your `/fee` endpoint requires a `type` parameter, add a
    ``TransactionForm.type`` attribute to the form. Polaris will detect the
    attribute's presence on the form and include it in `/fee` requests.

    The :attr:`amount` field is validated with the :meth:`clean_amount` function,
    which ensures the amount is within the bounds for the asset type.
    """

    def __init__(self, transaction: Transaction, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.transaction = transaction
        self.asset = transaction.asset
        self.decimal_places = self.asset.significant_decimals
        if transaction.kind == Transaction.KIND.deposit:
            self.min_amount = round(self.asset.deposit_min_amount, self.decimal_places)
            self.max_amount = round(self.asset.deposit_max_amount, self.decimal_places)
            self.min_default = (
                getattr(Asset, "_meta").get_field("deposit_min_amount").default
            )
            self.max_default = (
                getattr(Asset, "_meta").get_field("deposit_max_amount").default
            )
        else:
            self.min_amount = round(
                self.asset.withdrawal_min_amount, self.decimal_places
            )
            self.max_amount = round(
                self.asset.withdrawal_max_amount, self.decimal_places
            )
            self.min_default = (
                getattr(Asset, "_meta").get_field("withdrawal_min_amount").default
            )
            self.max_default = (
                getattr(Asset, "_meta").get_field("withdrawal_max_amount").default
            )

        # Re-initialize the 'amount' field now that we have all the parameters necessary
        self.fields["amount"].__init__(
            widget=forms.TextInput(
                attrs={
                    "class": "polaris-transaction-form-amount",
                    "inputmode": "decimal",
                    "symbol": self.asset.symbol,
                }
            ),
            min_value=self.min_amount,
            max_value=self.max_amount,
            decimal_places=self.decimal_places,
            label=_("Amount"),
            localize=True,
        )

        limit_str = ""
        if self.min_amount > self.min_default and self.max_amount < self.max_default:
            limit_str = f"({localize(self.min_amount)} - {localize(self.max_amount)})"
        elif self.min_amount > self.min_default:
            limit_str = _("(minimum: %s)") % localize(self.min_amount)
        elif self.max_amount < self.max_default:
            limit_str = _("(maximum: %s)") % localize(self.max_amount)

        if limit_str:
            self.fields["amount"].label += " " + limit_str

    amount = forms.DecimalField()

    def clean_amount(self):
        """Validate the provided amount of an asset."""
        amount = round(self.cleaned_data["amount"], self.decimal_places)
        if amount < self.min_amount:
            raise forms.ValidationError(
                _("The minimum amount is: %s")
                % localize(round(self.min_amount, self.decimal_places))
            )
        elif amount > self.max_amount:
            raise forms.ValidationError(
                _("The maximum amount is: %s")
                % localize(round(self.max_amount, self.decimal_places))
            )
        return amount



from typing import Dict

from polaris.locale.utils import (
    activate_lang_for_request,
    validate_or_use_default_language,
)
from polaris.utils import getLogger

from django.utils.translation import gettext as _
from rest_framework.decorators import api_view, renderer_classes
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer

from polaris import settings
from polaris.models import Asset
from polaris.utils import render_error_response
from polaris.integrations import (
    registered_fee_func,
    calculate_fee,
    registered_info_func,
)


logger = getLogger(__name__)


@api_view(["GET"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
def info(request: Request) -> Response:
    info_data = {
        "deposit": {},
        "withdraw": {},
        "fee": {"enabled": True, "authentication_required": True},
        "transactions": {"enabled": True, "authentication_required": True},
        "transaction": {"enabled": True, "authentication_required": True},
        "features": {"account_creation": True, "claimable_balances": True},
    }
    lang = validate_or_use_default_language(request.GET.get("lang"))
    activate_lang_for_request(lang)
    error_response = None
    for asset in Asset.objects.filter(sep6_enabled=True):
        error_response = populate_asset_info(request, asset, info_data, lang, False)
        if error_response:
            break
    if error_response:
        return error_response

    if "sep-38" not in settings.ACTIVE_SEPS:
        return Response(info_data)

    info_data["deposit-exchange"] = {}
    info_data["withdraw-exchange"] = {}
    for asset in Asset.objects.filter(sep6_enabled=True, sep38_enabled=True):
        error_response = populate_asset_info(request, asset, info_data, lang, True)
        if error_response:
            break
    if error_response:
        return error_response

    return Response(info_data)


def populate_asset_info(request, asset, info_data, lang, exchange=False):
    fields_and_types = registered_info_func(
        request=request,
        asset=asset,
        lang=lang,
        exchange=exchange,
    )
    try:
        validate_integration(fields_and_types)
    except ValueError as e:
        logger.error(f"info integration error: {str(e)}")
        return render_error_response(
            _("unable to process the request"), status_code=500
        )
    info_data["deposit-exchange" if exchange else "deposit"][
        asset.code
    ] = get_asset_info(asset, "deposit", fields_and_types.get("fields", {}))
    info_data["withdraw-exchange" if exchange else "withdraw"][
        asset.code
    ] = get_asset_info(asset, "withdrawal", fields_and_types.get("types", {}))


def validate_integration(fields_and_types: Dict):
    if not isinstance(fields_and_types, dict):
        raise ValueError("info integration must return a dictionary")
    elif not fields_and_types:
        # the anchor doesn't require additional arguments
        return
    fields = fields_and_types.get("fields")
    types = fields_and_types.get("types")
    if not set(fields_and_types.keys()).issubset({"fields", "types"}):
        raise ValueError("unexpected keys returned from info integration")
    if fields and not isinstance(fields, dict):
        raise ValueError("'fields' must be a dictionary")
    if types and not isinstance(types, dict):
        raise ValueError("'types' must be a dictionary")
    if fields:
        validate_fields(fields)
    for t, val in types.items():
        try:
            fields = val["fields"]
        except KeyError:
            raise ValueError(f"missing 'fields' key from {t}")
        if not isinstance(fields, dict):
            raise ValueError(f"'fields' key from {t} must be a dictionary")
        if len(val) != 1:
            raise ValueError(f"unexpected keys in {t} type")
        validate_fields(fields)


def validate_fields(fields: Dict):
    for val in fields.values():
        desc = val.get("description")
        optional = val.get("optional")
        choices = val.get("choices")
        if not desc:
            raise ValueError("'fields' dict must contain 'description'")
        if not set(val.keys()).issubset({"description", "optional", "choices"}):
            raise ValueError("unexpected keys in 'fields' dict")
        if not isinstance(desc, str):
            raise ValueError("'description' must be a string")
        if optional and not isinstance(optional, bool):
            raise ValueError("'optional' must be a boolean")
        if choices and not isinstance(choices, list):
            raise ValueError("'choices' must be a list")


def get_asset_info(asset: Asset, op_type: str, fields_or_types: Dict) -> Dict:
    if not getattr(asset, f"{op_type}_enabled"):
        return {"enabled": False}

    asset_info = {
        "enabled": True,
        "authentication_required": True,
    }
    min_amount_attr = f"{op_type}_min_amount"
    max_amount_attr = f"{op_type}_max_amount"
    min_amount = getattr(asset, min_amount_attr)
    max_amount = getattr(asset, max_amount_attr)
    if min_amount > getattr(Asset, "_meta").get_field(min_amount_attr).default:
        asset_info["min_amount"] = min_amount
    if max_amount < getattr(Asset, "_meta").get_field(max_amount_attr).default:
        asset_info["max_amount"] = max_amount
    if registered_fee_func is calculate_fee:
        # the anchor has not replaced the default fee function
        # so `fee_fixed` and `fee_percent` are still relevant.
        if getattr(asset, f"{op_type}_fee_fixed") is not None:
            asset_info["fee_fixed"] = getattr(asset, f"{op_type}_fee_fixed")
        if getattr(asset, f"{op_type}_fee_percent") is not None:
            asset_info["fee_percent"] = getattr(asset, f"{op_type}_fee_percent")

    if op_type == "deposit":
        asset_info["fields"] = fields_or_types
    else:
        asset_info["types"] = fields_or_types

    return asset_info



import pytz
from django.utils import timezone


class TimezoneMiddleware:
    """
    .. _timezones: https://docs.djangoproject.com/en/3.2/topics/i18n/timezones/

    Adapted from the Django documentation on timezones_. It checks for a
    ``"timezone"`` key stored in the request session and uses it when rendering
    content returned in the response.

    Polaris includes a ``timezone.js`` script that detects the users' UTC offset
    and sends it to the server, which stores a timezone with that offset in the
    user's session. This script is automatically loaded if using a template that
    inherits from ``base.html``.

    However, there is a limitation with this approach. For users's without exising
    sessions, which are identified using a browser cookie, Polaris cannot detect the
    user's timezone prior to rendering the first page of content. This means that
    dates and times shown to on the first page to a new user will be in the default
    timezone specified in your project's settings.

    That is why Django's documentation recommends that you simply ask the user what
    timezone they would like to use instead of attempting to detect it automatically.
    If this approach is taken, simply save the specified timezone in the user's session
    under the ``"timezone"`` key after adding this middleware.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        tzname = request.session.get("timezone")
        if tzname:
            timezone.activate(pytz.timezone(tzname))
        else:
            timezone.deactivate()
        return self.get_response(request)



"""This module defines the models used by Polaris."""
import datetime
import decimal
import secrets
import uuid
from base64 import urlsafe_b64encode as b64e, urlsafe_b64decode as b64d

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from django.core.exceptions import ValidationError
from django.core.validators import (
    MinLengthValidator,
    MinValueValidator,
    MaxValueValidator,
)
from django.db import models
from django.utils.encoding import force_bytes
from django.utils.translation import gettext_lazy as _
from model_utils import Choices
from model_utils.models import TimeStampedModel
from stellar_sdk.server_async import ServerAsync
from stellar_sdk.client.aiohttp_client import AiohttpClient
from stellar_sdk.exceptions import SdkError
from stellar_sdk.keypair import Keypair
from stellar_sdk.transaction_envelope import TransactionEnvelope

from polaris import settings

# Used for loading the distribution signers data onto an Asset obj
ASSET_DISTRIBUTION_ACCOUNT_MAP = {}


def utc_now():
    return datetime.datetime.now(datetime.timezone.utc)


class PolarisHeartbeat(models.Model):
    """
    Used as a locking mechanism to ensure that certain processes such as
    process_pending_deposits.py don't have more than 1 instance running
    at any given time. The last_heatbeat is a timestamp that is periodically
    updated by the process. If a process unexpectedly dies, another instance
    can check this value at startup and if its been too long since the last
    update, the lock is considered 'expired' and the new process can acquire
    it. This mechanism is an advisory lock and the locking logic is implemented
    at the application level.
    This value can also be used to create a 'health check' endpoint for the
    application
    Note: The application is expected to delete this key during a gracefully
    shutdown - see process_pending_deposits.py for an example
    """

    key = models.CharField(max_length=80, unique=True)
    last_heartbeat = models.DateTimeField(null=True, blank=True)

    objects = models.Manager()


class PolarisChoices(Choices):
    """A subclass to change the verbose default string representation"""

    def __repr__(self):
        return str(Choices)


class EncryptedTextField(models.TextField):
    """
    A custom field for ensuring its data is always encrypted at the DB
    layer and only decrypted by this object when in memory.

    Uses Fernet (https://cryptography.io/en/latest/fernet/) encryption,
    which relies on Django's SECRET_KEY setting for generating
    cryptographically secure keys.
    """

    @staticmethod
    def get_key(secret, salt):
        return b64e(
            PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=salt,
                iterations=100000,
                backend=default_backend(),
            ).derive(secret)
        )

    @classmethod
    def decrypt(cls, value):
        from django.conf import settings

        decoded = b64d(value.encode())
        salt, encrypted_value = decoded[:16], b64e(decoded[16:])
        key = cls.get_key(force_bytes(settings.SECRET_KEY), salt)
        return Fernet(key).decrypt(encrypted_value).decode()

    @classmethod
    def encrypt(cls, value):
        from django.conf import settings

        salt = secrets.token_bytes(16)
        key = cls.get_key(force_bytes(settings.SECRET_KEY), salt)
        encrypted_value = b64d(Fernet(key).encrypt(value.encode()))
        return b64e(b"%b%b" % (salt, encrypted_value)).decode()

    def from_db_value(self, value, *_args):
        if value is None:
            return value
        return self.decrypt(value)

    def get_db_prep_value(self, value, *_args, **_kwargs):
        if value is None:
            return value
        return self.encrypt(value)


class Asset(TimeStampedModel):
    code = models.TextField()
    """The asset code as defined on the Stellar network."""

    issuer = models.TextField(validators=[MinLengthValidator(56)])
    """The issuing Stellar account address."""

    significant_decimals = models.IntegerField(
        default=2, validators=[MinValueValidator(0), MaxValueValidator(7)]
    )
    """The number of decimal places Polaris should save when collecting input amounts"""

    # Deposit-related info
    deposit_enabled = models.BooleanField(default=True)
    """``True`` if deposit for this asset is supported."""

    deposit_fee_fixed = models.DecimalField(
        null=True, blank=True, max_digits=30, decimal_places=7
    )
    """
    Optional fixed (base) fee for deposit. In units of the deposited asset. 
    This is in addition to any ``fee_percent``. Omit if there is no fee or the fee 
    schedule is complex.
    """

    deposit_fee_percent = models.DecimalField(
        null=True,
        blank=True,
        max_digits=30,
        decimal_places=7,
        validators=[MinValueValidator(0), MaxValueValidator(100)],
    )
    """
    Optional percentage fee for deposit. In percentage points. This is in 
    addition to any ``fee_fixed``. Omit if there is no fee or the fee schedule
    is complex.
    """

    deposit_min_amount = models.DecimalField(
        default=0, blank=True, max_digits=30, decimal_places=7
    )
    """Optional minimum amount. No limit if not specified."""

    deposit_max_amount = models.DecimalField(
        default=decimal.MAX_EMAX, blank=True, max_digits=30, decimal_places=7
    )
    """Optional maximum amount. No limit if not specified."""

    # Withdrawal-related info
    withdrawal_enabled = models.BooleanField(default=True)
    """``True`` if withdrawal for this asset is supported."""

    withdrawal_fee_fixed = models.DecimalField(
        null=True, blank=True, max_digits=30, decimal_places=7
    )
    """
    Optional fixed (base) fee for withdraw. In units of the withdrawn asset. 
    This is in addition to any ``fee_percent``.
    """

    withdrawal_fee_percent = models.DecimalField(
        null=True,
        blank=True,
        max_digits=30,
        decimal_places=7,
        validators=[MinValueValidator(0), MaxValueValidator(100)],
    )
    """
    Optional percentage fee for withdraw in percentage points. This is in 
    addition to any ``fee_fixed``.
    """

    withdrawal_min_amount = models.DecimalField(
        default=0, blank=True, max_digits=30, decimal_places=7
    )
    """Optional minimum amount. No limit if not specified."""

    withdrawal_max_amount = models.DecimalField(
        default=decimal.MAX_EMAX, blank=True, max_digits=30, decimal_places=7
    )
    """Optional maximum amount. No limit if not specified."""

    send_fee_fixed = models.DecimalField(
        null=True, blank=True, max_digits=30, decimal_places=7
    )
    """
    Optional fixed (base) fee for sending this asset in units of this asset. 
    This is in addition to any ``send_fee_percent``. If null, ``fee_fixed`` will not
    be displayed in SEP31 /info response.
    """

    send_fee_percent = models.DecimalField(
        null=True,
        blank=True,
        max_digits=30,
        decimal_places=7,
        validators=[MinValueValidator(0), MaxValueValidator(100)],
    )
    """
    Optional percentage fee for sending this asset in percentage points. This is in 
    addition to any ``send_fee_fixed``. If null, ``fee_percent`` will not be displayed
    in SEP31 /info response.
    """

    send_min_amount = models.DecimalField(
        default=0, blank=True, max_digits=30, decimal_places=7
    )
    """Optional minimum amount. No limit if not specified."""

    send_max_amount = models.DecimalField(
        default=decimal.MAX_EMAX, blank=True, max_digits=30, decimal_places=7
    )
    """Optional maximum amount. No limit if not specified."""

    distribution_seed = EncryptedTextField(null=True, blank=True)
    """
    The distribution stellar account secret key.
    The value is stored in the database using Fernet symmetric encryption,
    and only decrypted when in the Asset object is in memory.
    """

    sep24_enabled = models.BooleanField(default=False)
    """`True` if this asset is transferable via SEP-24"""

    sep6_enabled = models.BooleanField(default=False)
    """`True` if this asset is transferable via SEP-6"""

    sep31_enabled = models.BooleanField(default=False)
    """`True` if this asset is transferable via SEP-31"""

    sep38_enabled = models.BooleanField(default=False)
    """`True` if this asset is exchangeable via SEP-38"""

    symbol = models.TextField(default="$")
    """The symbol used in HTML pages when displaying amounts of this asset"""

    objects = models.Manager()

    @property
    def distribution_account(self):
        """
        The Stellar public key derived from `Asset.distribution_seed`
        """
        if not self.distribution_seed:
            return None
        return Keypair.from_secret(str(self.distribution_seed)).public_key

    @property
    def asset_identification_format(self):
        """
        SEP-38 asset identification format
        """
        return f"stellar:{self.code}:{self.issuer}"

    def get_distribution_account_data(self, refresh=False):
        if refresh or self.distribution_account not in ASSET_DISTRIBUTION_ACCOUNT_MAP:
            account_json = (
                settings.HORIZON_SERVER.accounts()
                .account_id(self.distribution_account)
                .call()
            )
            ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account] = account_json
            return account_json
        else:
            return ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account]

    def get_distribution_account_signers(self, refresh=False):
        if refresh or self.distribution_account not in ASSET_DISTRIBUTION_ACCOUNT_MAP:
            account_json = self.get_distribution_account_data(refresh=refresh)
        else:
            account_json = ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account]
        return account_json["signers"]

    def get_distribution_account_thresholds(self, refresh=False):
        if refresh or self.distribution_account not in ASSET_DISTRIBUTION_ACCOUNT_MAP:
            account_json = self.get_distribution_account_data(refresh=refresh)
        else:
            account_json = ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account]
        return account_json["thresholds"]

    def get_distribution_account_master_signer(self, refresh=False):
        if refresh or self.distribution_account not in ASSET_DISTRIBUTION_ACCOUNT_MAP:
            account_json = self.get_distribution_account_data(refresh=refresh)
        else:
            account_json = ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account]
        master_signer = None
        for signer in account_json["signers"]:
            if signer["key"] == self.distribution_account:
                master_signer = signer
        return master_signer

    async def get_distribution_account_data_async(self, refresh=False):
        if refresh or self.distribution_account not in ASSET_DISTRIBUTION_ACCOUNT_MAP:
            async with ServerAsync(
                settings.HORIZON_URI, client=AiohttpClient()
            ) as server:
                account_json = (
                    await server.accounts().account_id(self.distribution_account).call()
                )
            ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account] = account_json
            return account_json
        else:
            return ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account]

    async def get_distributiion_account_signers_async(self, refresh=False):
        if refresh or self.distribution_account not in ASSET_DISTRIBUTION_ACCOUNT_MAP:
            account_json = await self.get_distribution_account_data_async(
                refresh=refresh
            )
        else:
            account_json = ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account]
        return account_json["signers"]

    async def get_distribution_account_thresholds_async(self, refresh=False):
        if refresh or self.distribution_account not in ASSET_DISTRIBUTION_ACCOUNT_MAP:
            account_json = await self.get_distribution_account_data_async(
                refresh=refresh
            )
        else:
            account_json = ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account]
        return account_json["thresholds"]

    async def get_distribution_account_master_signer_async(
        self, refresh=False, server=None
    ):
        if refresh or self.distribution_account not in ASSET_DISTRIBUTION_ACCOUNT_MAP:
            account_json = await self.get_distribution_account_data_async(
                refresh=refresh
            )
        else:
            account_json = ASSET_DISTRIBUTION_ACCOUNT_MAP[self.distribution_account]
        master_signer = None
        for signer in account_json["signers"]:
            if signer["key"] == self.distribution_account:
                master_signer = signer
        return master_signer

    class Meta:
        app_label = "polaris"

    def __str__(self):
        return f"{self.code} - issuer({self.issuer})"


def deserialize(value):
    """
    Validation function for Transaction.envelope_xdr
    """
    from polaris import settings

    try:
        TransactionEnvelope.from_xdr(value, settings.STELLAR_NETWORK_PASSPHRASE)
    except SdkError as e:
        raise ValidationError(
            _("Cannot decode envelope XDR for transaction: %(error)s"),
            params={"error": str(e)},
        )


class Transaction(models.Model):
    KIND = PolarisChoices(
        "deposit", "withdrawal", "send", "deposit-exchange", "withdrawal-exchange"
    )
    """Choices object for the kind of transaction"""

    status_to_message = {
        # SEP-6 & SEP-24
        "pending_anchor": _("processing"),
        "pending_trust": _("waiting for a trustline to be established"),
        "pending_user": _("waiting on user action"),
        "pending_user_transfer_start": _("waiting on the user to transfer funds"),
        "incomplete": _("incomplete"),
        "no_market": _("no market for the asset"),
        "too_small": _("the transaction amount is too small"),
        "too_large": _("the transaction amount is too big"),
        # SEP-31
        # messages are None because they are never displayed to user
        "pending_sender": None,
        "pending_receiver": None,
        "pending_transaction_info_update": None,
        "pending_customer_info_update": None,
        # Shared
        "completed": _("complete"),
        "error": _("error"),
        "pending_external": _("waiting on an external entity"),
        "pending_stellar": _("stellar is executing the transaction"),
    }

    SUBMISSION_STATUS = PolarisChoices(
        "not_ready",
        "ready",
        "processing",
        "pending",
        "pending_funding",
        "pending_trust",
        "blocked",
        "unblocked",
        "completed",
        "failed",
    )
    """ 
        Submission Statuses

    * **not_ready**
        used until a transaction is returned from RailsIntegration.poll_pending_deposits()
        and determined by check_account to be ready for submission to the Stellar Network
    
    * **ready**
        used when the transaction has been processed by the check_account task and verified
        that the transaction is ready to be submitted to the Stellar Network

    * **processing**
        used when Polaris is submitting the transaction to Stellar. Note that up to two 
        transactions could be submitted for this Transaction object, one for creating the 
        account if it doesn't exist, and the other for sending the deposit payment.

    * **pending**
        used when the transaction has been passed to 
        CustodyIntegration.create_destination_account() or 
        CustodyIntegration.submit_deposit_transaction() but a `TransactionSubmissionPending` 
        exception was raised in the most-recent invocation, and a SIGINT or SIGTERM was 
        sent, preventing Polaris from submitting again.
    
    * **pending_trust**
        used when the transaction destination account does not yet have a trustline
    
    * **blocked**
        used when the transaction has been passed to 
        CustodyIntegration.create_destination_account()
        or CustodyIntegration.submit_deposit_transaction() but a `TransactionSubmissionBlocked` 
        exception was raised in the most-recent invocation. Polaris will simply move to the 
        next transaction.
    
    * **unblocked**
        Similar to READY, but indicates that the transaction was previously blocked.
    
    * **completed**
        used when a transaction has been successfully submitted to the Stellar network

    * **failed**
        used when a transaction has been passed to 
        CustodyIntegration.submit_deposit_transaction() but a `TransactionSubmissionFailed`
        exception was raised
    """

    STATUS = PolarisChoices(*list(status_to_message.keys()))

    MEMO_TYPES = PolarisChoices("text", "id", "hash")
    """Type for the ``memo``. Can be either `hash`, `id`, or `text`"""

    PROTOCOL = PolarisChoices("sep6", "sep24", "sep31")
    """Values for `protocol` column"""

    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    """Unique, anchor-generated id for the deposit/withdrawal."""

    paging_token = models.TextField(null=True, blank=True)
    """The token to be used as a cursor for querying before or after this transaction"""

    stellar_account = models.TextField(validators=[MinLengthValidator(1)])
    """
    The Stellar (G...) account authenticated via SEP-10 that initiated this transaction.
    Note that if ``Transaction.muxed_account`` is not null, this column's value is 
    derived from the muxed account.
    """

    muxed_account = models.TextField(null=True, blank=True)
    """
    The muxed (M...) account authenticated via SEP-10 that initiated this transaction.
    If this column value is not null, ``Transaction.stellar_account`` is derived from
    this value and ``Transaction.account_memo`` will be null.
    """

    account_memo = models.PositiveIntegerField(null=True, blank=True)
    """
    The ID (64-bit integer) memo identifying the user of the shared Stellar account 
    authenticated via SEP-10 that initiated this transaction. If this column value
    is not null, ``Transaction.muxed_account`` will be null.
    """

    asset = models.ForeignKey("Asset", on_delete=models.CASCADE)
    """The Django foreign key to the associated :class:`Asset`"""

    quote = models.ForeignKey("Quote", null=True, blank=True, on_delete=models.CASCADE)

    # These fields can be shown through an API:
    kind = models.CharField(choices=KIND, default=KIND.deposit, max_length=20)
    """The character field for the available ``KIND`` choices."""

    submission_status = models.CharField(
        choices=SUBMISSION_STATUS, default=SUBMISSION_STATUS.not_ready, max_length=31
    )

    status = models.CharField(
        choices=STATUS, default=STATUS.pending_external, max_length=31
    )
    """
    Choices field for processing status of deposit, withdrawal, & send.
    
    SEP-6 & SEP-24 Statuses:

    * **completed**
        
        deposit/withdrawal fully completed
    * **pending_external**
    
        deposit/withdrawal has been submitted to external 
        network, but is not yet confirmed. This is the status when waiting on 
        Bitcoin or other external crypto network to complete a transaction, or 
        when waiting on a bank transfer.
    * **pending_anchor**
    
        deposit/withdrawal is being processed internally by anchor.
    * **pending_stellar**
    
        deposit/withdrawal operation has been submitted to Stellar network, but 
        is not yet confirmed.
    * **pending_trust**
    
        the user must add a trust-line for the asset for the deposit to complete.
    * **pending_user**
    
        the user must take additional action before the deposit / withdrawal can 
        complete.
    * **pending_user_transfer_start**
    
        the user has not yet initiated their transfer to the anchor. This is the 
        necessary first step in any deposit or withdrawal flow.

    * **incomplete**
    
        there is not yet enough information for this transaction to be initiated. 
        Perhaps the user has not yet entered necessary info in an interactive flow.
    * **no_market**
    
        could not complete deposit because no satisfactory asset/XLM market 
        was available to create the account.
    * **too_small**
    
        deposit/withdrawal size less than min_amount.
    * **too_large**
    
        deposit/withdrawal size exceeded max_amount.
    * **error**
    
        catch-all for any error not enumerated above.
        
    SEP-31 Statuses:
    
    * **pending_sender**
    
        awaiting payment to be initiated by sending anchor.
    * **pending_stellar**
    
        transaction has been submitted to Stellar network, but is not yet confirmed.
    * **pending_transaction_info_update**
    
        transaction details must be updated to successfully execute transaction off-chain
    * **pending_customer_info_update**
    
        customer (SEP-12) information must be updated to facilitate transactions
    * **pending_receiver**
    
        payment is being processed by the receiving anchor.
    * **pending_external**
    
        payment has been submitted to external network, but is not yet confirmed.
    * **completed**
    
        deposit/withdrawal fully completed.
    * **error**
    
        catch-all for any error not enumerated above.
    """

    status_eta = models.IntegerField(null=True, blank=True)
    """(optional) Estimated number of seconds until a status change is expected."""

    status_message = models.TextField(null=True, blank=True)
    """A message stored in association to the current status for debugging"""

    stellar_transaction_id = models.TextField(null=True, blank=True)
    """
    transaction_id on Stellar network of the transfer that either completed
    the deposit or started the withdrawal.
    """

    external_transaction_id = models.TextField(null=True, blank=True)
    """
    (optional) ID of transaction on external network that either started 
    the deposit or completed the withdrawal.
    """

    amount_in = models.DecimalField(
        null=True, blank=True, max_digits=30, decimal_places=7
    )
    """
    Amount received by anchor at start of transaction as a string with up 
    to 7 decimals. Excludes any fees charged before the anchor received the 
    funds.
    """

    amount_expected = models.DecimalField(
        null=True, blank=True, max_digits=30, decimal_places=7
    )
    """
    Amount the client specified would be sent to the anchor at the start of
    a transaction. Note that ``Transaction.amount_in`` can differ from this field 
    after funds have been received. Until then, the fields will match. This field
    makes it possible to check if the amount sent to the anchor matches the amount
    the client initially specified in an API request or form.
    """

    amount_out = models.DecimalField(
        null=True, blank=True, max_digits=30, decimal_places=7
    )
    """
    Amount sent by anchor to user at end of transaction as a string with up to
    7 decimals. Excludes amount converted to XLM to fund account and any 
    external fees.
    """

    amount_fee = models.DecimalField(
        null=True, blank=True, max_digits=30, decimal_places=7
    )
    """Amount of fee charged by anchor."""

    fee_asset = models.TextField(null=True, blank=True)
    """
    The string representing the asset in which the fee is charged. The string
    must be formatted using SEP-38's Asset Identification Format, and is only
    necessary for transactions using different on and off-chain assets.
    """

    queue = models.TextField(null=True, blank=True)
    """The queue that this transaction is currently in"""

    queued_at = models.DateTimeField(null=True, blank=True)
    """The time when this transaction was queued"""

    started_at = models.DateTimeField(default=utc_now)
    """Start date and time of transaction."""

    completed_at = models.DateTimeField(null=True, blank=True)
    """
    Completion date and time of transaction. Assigned null for in-progress 
    transactions.
    """

    from_address = models.TextField(
        null=True, blank=True
    )  # Using from_address since `from` is a reserved keyword
    """Sent from address, perhaps BTC, IBAN, or bank account."""

    to_address = models.TextField(
        null=True, blank=True
    )  # Using to_address for naming consistency
    """
    Sent to address (perhaps BTC, IBAN, or bank account in the case of a 
    withdrawal or send, Stellar or muxed address in the case of a deposit).
    """

    required_info_updates = models.TextField(null=True, blank=True)
    """
    (SEP31) (optional) A set of fields that require an update from the sender, 
    in the same format as described in /info.
    """

    required_info_message = models.TextField(null=True, blank=True)
    """
    (SEP31) (optional) A human readable message indicating any errors that 
    require updated information from the sender
    """

    memo = models.TextField(null=True, blank=True)
    """
    (optional) Value of memo to attach to transaction, for hash this should
    be base64-encoded.
    """

    memo_type = models.CharField(
        choices=MEMO_TYPES, default=MEMO_TYPES.text, max_length=10
    )
    """
    (optional) Type of memo that anchor should attach to the Stellar payment 
    transaction, one of text, id or hash.
    """

    receiving_anchor_account = models.TextField(null=True, blank=True)
    """
    Stellar account to send payment or withdrawal funds to
    """

    refunded = models.BooleanField(default=False)
    """True if the transaction was refunded, false otherwise."""

    protocol = models.CharField(choices=PROTOCOL, null=True, max_length=5, blank=True)
    """Either 'sep6', 'sep24', or 'sep31'"""

    pending_signatures = models.BooleanField(default=False)
    """
    Boolean for whether or not non-Polaris signatures are needed for this 
    transaction's envelope.
    """

    envelope_xdr = models.TextField(validators=[deserialize], null=True, blank=True)
    """
    The base64-encoded XDR blob that can be deserialized to inspect and sign 
    the encoded transaction.
    """

    channel_seed = EncryptedTextField(null=True, blank=True)
    """
    A keypair of the account used when sending SEP-6 or SEP-24 deposit 
    transactions to Transaction.to_address, if present. 
    This is only used for transactions requiring signatures Polaris cannot
    add itself.
    """

    claimable_balance_supported = models.BooleanField(default=False)
    """
    claimable_balance_supported is a boolean to indicate if the wallet supports the SEP24
    requirements for handeling claimable balance deposits.
    """

    claimable_balance_id = models.TextField(null=True, blank=True)
    """
    The ID of the claimable balance used to send funds to the user. This column will be
    ``None`` if ``claimable_balance_supported`` is ``False`` or if the transaction has
    not yet been submitted to the Stellar network.
    """

    more_info_url = models.TextField(null=True, blank=True)
    """
    A URL that is opened by wallets after the interactive flow is complete. It can include
    banking information for users to start deposits, the status of the transaction, or any
    other information the user might need to know about the transaction.
    """

    on_change_callback = models.TextField(null=True, blank=True)
    """
    A URL that the anchor should POST a JSON message to when the status property of the
    transaction created as a result of this request changes.
    """

    pending_execution_attempt = models.BooleanField(default=False)
    """
    An internal column used to ensure transactions are not retrieved from the database 
    and executed by different processes running the same command, specifically 
    process_pending_deposits and execute_outgoing_transactions.
    """

    client_domain = models.TextField(null=True, blank=True)
    """
    The hostname of the client application that requested this transaction on behalf of
    the user. The SIGNING_KEY on `https://client_domain/.well-known/stellar.toml` signed 
    the challenge transaction used to obtain the authentication token necessary to 
    request this transaction, effectively allowing requests including the authentication 
    token to be attributed to it.
    """

    objects = models.Manager()

    @property
    def asset_name(self):
        return self.asset.code + ":" + self.asset.issuer

    @property
    def message(self):
        """
        Human readable explanation of transaction status
        """
        return self.status_to_message[str(self.status)]

    @property
    def channel_account(self):
        if not self.channel_seed:
            return None
        return Keypair.from_secret(str(self.channel_seed)).public_key

    class Meta:
        ordering = ("-started_at",)
        app_label = "polaris"


class Quote(models.Model):
    """
    Quote objects represent either firm or indicative quotes requested by the client
    application and provided by the anchor. Quote objects will be assigned to the
    Transaction.quote column by Polaris when requested via a SEP-6 or SEP-31 request.
    Anchors must create their own Quote objects when facilitating a SEP-24 transaction.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    """
    The unique ID for the quote.
    """

    stellar_account = models.TextField()
    """
    The Stellar (G...) account authenticated via SEP-10 when this Quote was created.
    Note that if ``Quote.muxed_account`` is not null, this column's value is 
    derived from the muxed account. 
    """

    account_memo = models.PositiveIntegerField(null=True, blank=True)
    """
    The ID (64-bit integer) memo identifying the user of the shared Stellar account 
    authenticated via SEP-10 when this Quote was created. If this column value
    is not null, ``Quote.muxed_account`` will be null.
    """

    muxed_account = models.TextField(null=True, blank=True)
    """
    The muxed (M...) account authenticated via SEP-10 when this Quote was created.
    If this column value is not null, ``Quote.stellar_account`` is derived from
    this value and ``Quote.account_memo`` will be null.
    """

    TYPE = PolarisChoices("firm", "indicative")
    """
    Choices for type.
    """

    type = models.TextField(choices=TYPE)
    """
    The type of quote. Firm quotes have a non-null price and expiration, indicative quotes 
    may have a null price and expiration.
    """

    sell_asset = models.TextField()
    """
    The asset the client would like to sell. Ex. USDC:G..., iso4217:ARS
    """

    buy_asset = models.TextField()
    """
    The asset the client would like to receive for some amount of sell_asset.
    """

    sell_amount = models.DecimalField(max_digits=30, decimal_places=7)
    """
    The amount of sell_asset the client would exchange for buy_asset.
    """

    buy_amount = models.DecimalField(
        null=True, blank=True, max_digits=30, decimal_places=7
    )
    """
    The amount of buy_asset the client would like to purchase with sell_asset.
    """

    price = models.DecimalField(null=True, blank=True, max_digits=30, decimal_places=7)
    """
    The price offered by the anchor for one unit of buy_asset in terms of sell_asset.
    """

    expires_at = models.DateTimeField(null=True, blank=True)
    """
    The expiration time of the quote. Null if type is Quote.TYPE.indicative.
    """

    sell_delivery_method = models.ForeignKey(
        "DeliveryMethod",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name="+",
    )
    """
    One of the name values specified by the sell_delivery_methods array.
    """

    buy_delivery_method = models.ForeignKey(
        "DeliveryMethod",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name="+",
    )
    """
    One of the name values specified by the buy_delivery_methods array.
    """

    country_code = models.TextField(null=True, blank=True)
    """
    The ISO 3166-1 alpha-3 code of the user's current address. 
    """

    requested_expire_after = models.DateTimeField(null=True, blank=True)
    """
    The requested expiration date from the client.
    """

    objects = models.Manager()


class OffChainAsset(models.Model):
    """
    Off-chain assets represent the asset being exchanged with the Stellar asset. Each
    off-chain asset has a set of delivery methods by which the user can provide funds to
    the anchor and by which the anchor can deliver funds to the user.
    """

    scheme = models.TextField()
    """
    The scheme of the off-chain asset as defined by SEP-38's Asset Identification Format.
    """

    identifier = models.TextField()
    """
    The identifier of the off-chain asset as defined by SEP-38's Asset Identification Format.
    """

    significant_decimals = models.PositiveIntegerField(default=2)
    """
    The number of decimal places Polaris should preserve when collecting & calculating amounts.
    """

    country_codes = models.TextField(null=True, blank=True)
    """
    A comma-separated list of ISO 3166-1 alpha-3 codes of the countries where the anchor 
    supports delivery of this asset.
    """

    delivery_methods = models.ManyToManyField("DeliveryMethod")
    """
    The list of delivery methods support for collecting and receiving this asset
    """

    symbol = models.TextField(null=True, blank=True)
    """
    The symbol to use when displaying amounts of this asset
    """

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["scheme", "identifier"], name="offchain_unique_index"
            )
        ]

    objects = models.Manager()

    @property
    def asset_identification_format(self):
        return f"{self.scheme}:{self.identifier}"


class DeliveryMethod(models.Model):
    """
    Delivery methods are the supported means of payment from the user to the anchor and from
    the anchor to the user. For example, an anchor may have retail stores that accept cash
    drop-off and pick-up, or only accept debit or credit card payments. The method used by
    the anchor to collect or deliver funds to the user may affect the rate or fees charged
    for facilitating the transaction.
    """

    TYPE = PolarisChoices("buy", "sell")
    """
    The types of delivery methods.
    """

    type = models.TextField(choices=TYPE)
    """
    The type of delivery method. Sell methods describe how a client can deliver funds to the 
    anchor. Buy methods describe how a client can receive or collect funds from the anchor.
    """

    name = models.TextField()
    """
    The name of the delivery method, to be used in SEP-38 request and response bodies.
    """

    description = models.TextField()
    """
    The human-readable description of the deliver method, to be used in SEP-38 
    response bodies.
    """

    objects = models.Manager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["name", "type"], name="deliverymethod_unique_index"
            )
        ]


class ExchangePair(models.Model):
    """
    Exchange pairs consist of an off-chain and on-chain asset that can be exchanged.
    Specifically, one of these assets can be sold by the client (sell_asset) and the
    other is bought by the client (buy_asset). ExchangePairs cannot consist of two
    off-chain assets or two on-chain assets. Note that two exchange pair objects must
    be created if each asset can be bought or sold for the other.
    """

    buy_asset = models.TextField()
    """
    The asset the client can purchase with sell_asset using SEP-38's Asset 
    Identification Format.
    """

    sell_asset = models.TextField()
    """
    The asset the client can provide in exchange for buy_asset using SEP-38's
    Asset Identification Format.
    """

    objects = models.Manager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["buy_asset", "sell_asset"], name="exchangepair_unique_index"
            )
        ]



import json
from datetime import timezone

from rest_framework import serializers

from polaris.models import Transaction
from polaris.settings import DATETIME_FORMAT


class SEP31TransactionSerializer(serializers.ModelSerializer):
    id = serializers.CharField()
    stellar_account_id = serializers.SerializerMethodField()
    stellar_memo = serializers.SerializerMethodField()
    stellar_memo_type = serializers.SerializerMethodField()
    required_info_message = serializers.SerializerMethodField()
    required_info_updates = serializers.SerializerMethodField()
    started_at = serializers.DateTimeField(
        format=DATETIME_FORMAT, default_timezone=timezone.utc
    )
    completed_at = serializers.DateTimeField(
        format=DATETIME_FORMAT, default_timezone=timezone.utc
    )

    @staticmethod
    def get_stellar_account_id(instance):
        return instance.receiving_anchor_account

    @staticmethod
    def get_stellar_memo(instance):
        return instance.memo

    @staticmethod
    def get_stellar_memo_type(instance):
        return instance.memo_type

    @staticmethod
    def get_required_info_message(instance):
        return instance.required_info_message

    @staticmethod
    def get_required_info_updates(instance):
        if instance.required_info_updates:
            return json.loads(instance.required_info_updates)
        else:
            return None

    def to_representation(self, instance):
        data = super().to_representation(instance)
        self._round_decimals(data, instance)
        if instance.quote:
            data["amount_in_asset"] = instance.quote.sell_asset
            data["amount_out_asset"] = instance.quote.buy_asset
            if instance.fee_asset:
                data["amount_fee_asset"] = instance.fee_asset
        return data

    @staticmethod
    def _round_decimals(data, instance):
        """
        Rounds each decimal field to instance.asset.significant_decimals.

        Note that this requires an additional database query for the asset.
        If this serializer was initialized to serialize many instances, this
        function will be called for each instance unless ``same_asset: True``
        is included as a key-value pair in self.context.

        If the transactions to be serialized are for multiple assets, split
        the calls to this serializer by asset.
        """
        asset = instance.asset
        for field in ["amount_in", "amount_out", "amount_fee"]:
            if getattr(instance, field) is None:
                continue
            data[field] = str(
                round(getattr(instance, field), asset.significant_decimals)
            )

    class Meta:
        model = Transaction
        fields = [
            "id",
            "status",
            "status_eta",
            "amount_in",
            "amount_out",
            "amount_fee",
            "started_at",
            "completed_at",
            "stellar_transaction_id",
            "external_transaction_id",
            "refunded",
            # fields with getter methods
            "stellar_account_id",
            "stellar_memo",
            "stellar_memo_type",
            "required_info_updates",
            "required_info_message",
        ]



from typing import Dict, Optional, List
from decimal import Decimal

from django import forms
from django.http import QueryDict
from rest_framework.request import Request

from polaris.models import Transaction, Asset
from polaris.integrations.forms import TransactionForm
from polaris.templates import Template
from polaris.sep10.token import SEP10Token


class DepositIntegration:
    """
    The container class for deposit integration functions.

    Subclasses must be registered with Polaris by passing it to
    :func:`polaris.integrations.register_integrations`.
    """

    def after_deposit(self, transaction: Transaction, *args: List, **kwargs: Dict):
        """
        Use this function to perform any post-processing of `transaction` after
        its been executed on the Stellar network. This could include actions such
        as updating other django models in your project or emailing users about
        completed deposits. Overriding this function is only required for anchors
        with multi-signature distribution accounts.

        If the transacted asset's distribution account requires multiple signatures,
        ``transaction.channel_account`` was created on Stellar when Polaris made a call
        to ``create_channel_account()`` and was used as the transaction-level source
        account when submitting to Stellar. This temporary account holds a minimum a
        XLM reserve balance that must be merged back to persistent account owned by
        the anchor. Generally, the destination account for the merge operation will be
        the same account that created and funded the channel account.

        :param transaction: a ``Transaction`` that was executed on the
            Stellar network
        """
        raise NotImplementedError()

    def form_for_transaction(
        self,
        request: Request,
        transaction: Transaction,
        post_data: Optional[QueryDict] = None,
        amount: Optional[Decimal] = None,
        *args: List,
        **kwargs: Dict
    ) -> Optional[forms.Form]:
        """
        This function should return the next form to render for the user given the
        state of the interactive flow.

        For example, this function could return an instance of a ``TransactionForm``
        subclass. Once the form is submitted, Polaris will detect the form used
        is a ``TransactionForm`` subclass and update ``transaction.amount_in``
        with the amount specified in form.

        If `post_data` is passed, it must be used to initialize the form returned so
        Polaris can validate the data submitted. If `amount` is passed, it can be used
        to pre-populate a ``TransactionForm`` amount field to improve the user
        experience.
        ::

            def form_for_transaction(self, transaction, post_data = None, amount = None):
                if transaction.amount_in:
                    return
                elif post_data:
                    return TransactionForm(transaction, post_data)
                else:
                    return TransactionForm(transaction, initial={"amount": amount})

        After a form is submitted and validated, Polaris will call
        ``DepositIntegration.after_form_validation`` with the populated
        form and transaction. This is where developers should update their own
        state-tracking constructs or do any processing with the data submitted
        in the form.

        Finally, Polaris will call this function again to check if there is
        another form that needs to be rendered to the user. If you are
        collecting KYC data, you can return a ``forms.Form`` with the fields you
        need.

        This loop of submitting a form, validating & processing it, and checking
        for the next form will continue until this function returns ``None``.

        When that happens, Polaris will check if ``content_for_template()`` also
        returns ``None``. If that is the case, Polaris assumes the anchor is finished
        collecting information and will update the Transaction status to
        ``pending_user_transfer_start``.

        If ``content_for_template()`` returns a dictionary, Polaris will serve a page
        `without` a form. Anchors should do this when the user needs to take some action
        in order to continue, such as confirming their email address. Once the user is
        confirmed, ``form_for_transaction()`` should return the next form.

        :param request: the ``rest_framework.request.Request`` object
        :param transaction: the ``Transaction`` database object
        :param post_data: the data sent in the POST request as a dictionary
        :param amount: a ``Decimal`` object the wallet may pass in the GET request.
            Use it to pre-populate your TransactionForm along with any SEP-9_
            parameters.
        :return: a dictionary containing various pieces of information to use
            when rendering the next page.
        """
        if transaction.amount_in:
            # we've collected transaction info
            # and don't implement KYC by default
            return

        if post_data:
            return TransactionForm(transaction, post_data)
        else:
            return TransactionForm(transaction, initial={"amount": amount})

    def content_for_template(
        self,
        request: Request,
        template: Template,
        form: Optional[forms.Form] = None,
        transaction: Optional[Transaction] = None,
        *args: List,
        **kwargs: Dict
    ) -> Optional[Dict]:
        """
        .. _`widget attributes`: https://docs.djangoproject.com/en/3.2/ref/forms/widgets/#styling-widget-instances
        .. _`Django template variables`: https://docs.djangoproject.com/en/3.2/ref/templates/language/#variables
        .. _`SEP-38 Asset Identification Format`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0038.md#asset-identification-format

        Return a dictionary containing the `Django template variables`_ to be passed to
        the template rendered.

        The anchor may also pass a special key, `template_name`, which should be a file
        path relative your Django app's `/templates` directory. Polaris will render the
        template specified by this key to the user instead of the default templates defined
        below. Note that all of the `Django template variables`_ defined below will still
        be passed to the template specified.

        Polaris will pass one of the following ``polaris.templates.Template`` values to
        indicate the default template Polaris will use.

        ``Template.DEPOSIT``

            The template used for deposit flows

        ``Template.WITHDRAW``

            The template used for withdraw flows

        ``Template.MORE_INFO``

            The template used to show transaction details

        The `form` parameter will always be ``None`` when `template` is
        ``Template.MORE_INFO`` since that page does not display form content.

        If `form` is ``None`` and `template_name` is **not** ``Template.MORE_INFO``,
        returning ``None`` will signal to Polaris that the anchor is done collecting
        information for the transaction. Returning content will signal to Polaris that
        the user needs to take some action before receiving the next form, such as
        confirming their email. In this case, make sure to return an appropriate
        `guidance` message or return a custom `template_name`.

        Using this function, anchors pass key-value pairs to the template being rendered.
        Some of these key-value pairs are used by Polaris, but anchors are allowed and
        encouraged to extend Polaris' templates and pass custom key-value pairs.
        ::

            def content_for_template(template_name, form=None, transaction=None):
                ...
                return {
                    "title": "Deposit Transaction Form",
                    "guidance": "Please enter the amount you would like to deposit.",
                    "show_fee_table": True,
                    "symbol": "$",
                    "icon_label": "Stellar Development Foundation",
                    "icon_path": "images/company-icon.png",
                    # custom field passed by the anchor
                    "username": "John.Doe"
                }

        Below are all the keys passed to the template rendered. If the dictionary returned has
        the same key, the default value Polaris uses will be overwritten.

        For ``Template.DEPOSIT`` and ``Template.WITHDRAW``:

        ``form``

            The ``django.forms.Form`` instance returned from ``form_for_transaction()``.

        ``post_url``

            The URL to make the POST request containing the form data to.

        ``operation``

            Either `deposit` or `withdraw`.

        ``asset``

            The ``polaris.models.Asset`` object of the Stellar asset being transacted.

        ``use_fee_endpoint``

            A boolean indicating whether or not Polaris should use the ``GET /fee``
            endpoint when calculating fees and rendering the amounts on the page.

        ``org_logo_url``

            A URL for the default logo to render if the anchor has not specified their own
            via `icon_path`.

        ``additive_fees_enabled``

            A boolean indicating whether or not to add fees to the amount entered in
            ``TransactionForm`` amount fields. ``False`` by default, meaning fees are
            subtracted from the amounts entered.

        ``title``

            The browser tab's title.

        ``guidance``

            A text message displayed on the page that should help guide the user to take
            the appropriate action(s).

        ``icon_label``

            The label for the image rendered on the page specified by ``"icon_path"``.

        ``icon_path``

            The relative file path to the image you would like to use as the company icon
            in the UI. The file path should be relative to your Django app's `/static`
            directory. If `icon_path` is not present, Polaris will use the image specified by
            your TOML integration function's ``ORG_LOGO`` key. If neither are present,
            Polaris will use its default image. All images will be rendered in a 100 x 150px
            sized box as defined by the default stylesheet.

        ``show_fee_table``

            A boolean for whether the fee table in the default template should be visible
            on the page rendered to the user. This table is hidden by default unless
            a ``TransactionForm`` is returned from ``form_for_transaction()``, in which
            case the fee table will be displayed. If the anchor instructs Polaris to display
            the fee table but a ``TransactionForm`` is not present on the page, the anchor is
            responsible for updating the fee table with the appropriate values. This is
            useful when the anchor is collecting the amount of an off-chain asset, since
            ``TransactionForm`` assumes the amount collected is for an on-chain asset.

        ``symbol``

            The character string that precedes the amounts shown on the fee table. It defaults
            to the Stellar ``Asset.symbol``. Note that the symbol used in input fields must
            be passed separately using the field's `widget attributes`_.

        For ``Template.MORE_INFO``

        ``tx_json``

            A JSON-serialized string matching the schema returned from `GET /transaction`

        ``amount_in_asset``

            The string representation of the asset given to the anchor by the user,
            formatted using `SEP-38 Asset Identification Format`_.

        ``amount_out_asset``

            The string representation of the asset sent from the anchor to the user,
            formatted using `SEP-38 Asset Identification Format`_.

        ``amount_in``

            A string containing the amount to be displayed on the page as `Amount Sent`

        ``amount_out``

            A string containing the amount to be displayed on the page as `Amount Received`

        ``amount_fee``

            A string containing the amount to be displayed on the page as `Fee`

        ``amount_in_symbol``

            ``Asset.symbol`` or ``OffChainAsset.symbol``, depending on whether or not
            asset sent to the anchor is on or off chain. If ``Transaction.quote`` is
            null, the value will always match ``Asset.symbol``.

        ``amount_fee_symbol``

            ``Asset.symbol`` or ``OffChainAsset.symbol``, depending on the value of
            ``Transaction.fee_asset``. If ``Transaction.quote`` is null, the value will
            always be ``Asset.symbol``.

        ``amount_out_symbol``

            ``Asset.symbol`` or ``OffChainAsset.symbol``, depending on whether or not
            asset sent by the anchor is on or off chain. If ``Transaction.quote`` is
            null, the value will always match ``Asset.symbol``.

        ``amount_in_significant_decimals``

            The number of decimals to display for amounts of ``Transaction.amount_in``.
            Derived from ``Asset.significant_decimals`` or ``OffChainAsset.decimals``.
            If ``Transaction.quote`` is null, the value will always match
            ``Asset.significant_decimals``.

        ``amount_fee_significant_decimals``

            The number of decimals to display for amounts of ``Transaction.amount_fee``.
            Derived from ``Asset.significant_decimals`` or ``OffChainAsset.decimals``,
            depending on the value of ``Transaction.fee_asset``.

        ``amount_out_significant_decimals``

            The number of decimals to display for amounts of ``Transaction.amount_out``.
            Derived from ``Asset.significant_decimals`` or ``OffChainAsset.decimals``.
            If ``Transaction.quote`` is null, the value will always match
            ``Asset.significant_decimals``.

        ``transaction``

            The ``polaris.models.Transaction`` object representing the transaction.

        ``asset``

            The ``polaris.models.Asset`` object representing the asset.

        ``offchain_asset``

            The ``OffChainAsset`` object used in the ``Transaction.quote``, if present.

        ``price``

            ``Transaction.quote.price``, if present.

        ``price_inversion``

            ``1 / Transaction.quote.price``, if ``price`` is present. The default
            `more_info.html` template uses this number for displaying exchange rates
            when quotes are used.

        ``price_inversion_significant_decimals``

            The number of decimals to display for exchange rates. Polaris calculates
            this to ensure the rate displayed is always correct.

        ``exchange_amount``

            If ``Transaction.quote`` is not ``None``, ``exchange_amount`` is the
            value of ``Transaction.amount_out`` expressed in units of
            ``Transaction.amount_in``.

        ``exchanged_amount``

            If ``Transaction.quote`` is not ``None``, ``exchanged_amount`` is the
            value of ``Transaction.amount_in`` expressed in units of
            ``Transaction.amount_out``.

        :param request: a ``rest_framework.request.Request`` instance
        :param template: a ``polaris.templates.Template`` enum value
            for the template to be rendered in the response
        :param form: the form to be rendered in the template
        :param transaction: the transaction being processed
        """
        raise NotImplementedError()

    def after_form_validation(
        self,
        request: Request,
        form: forms.Form,
        transaction: Transaction,
        *args: List,
        **kwargs: Dict
    ):
        """
        Use this function to process the data collected with `form` and to update
        the state of the interactive flow so that the next call to
        ``DepositIntegration.form_for_transaction()`` returns the next form to render
        to the user.

        If you need to store some data to determine which form to return next when
        ``DepositIntegration.form_for_transaction`` is called, store this
        data in a model not used by Polaris.

        If `form` is the last form to be served to the user, Polaris will update the
        transaction status to ``pending_user_transfer_start``, indicating that the
        anchor is waiting for the user to deliver off-chain funds to the anchor. If
        the KYC information collected is still being verified, update the
        ``Transaction.status`` column to ``pending_anchor`` here. Make sure to save
        this change to the database before returning. In this case Polaris will
        detect the status change and will not update the status again. Polaris will
        wait until the anchor changes the transaction's status to
        ``pending_user_transfer_start`` before including the transaction in calls to
        ``DepositIntegration.poll_pending_deposits()``.

        If the user is requesting a deposit or withdrawal of a Stellar asset in
        exchange for different off-chain asset, such as requesting a deposit of
        USDC using fiat mexican pesos, the anchor must assign a ``Quote`` object
        to ``Transaction.quote`` before the end of the interactive flow. Polaris
        will check for a ``Quote`` object on the transaction and adjust the UI
        of the ``MORE_INFO`` template to display the exchange rate and other
        exchange-related information. ``Transaction.fee_asset`` also must be
        populated with the asset in which fees will be collected, formatted
        using `SEP-38 Asset Identification Format`_.

        :param request: the ``rest_framework.request.Request`` object
        :param form: the completed ``forms.Form`` submitted by the user
        :param transaction: the ``Transaction`` database object
        """
        raise NotImplementedError()

    def interactive_url(
        self,
        request: Request,
        transaction: Transaction,
        asset: Asset,
        amount: Optional[Decimal],
        callback: Optional[str],
        lang: Optional[str],
        *args: List,
        **kwargs: Dict
    ) -> Optional[str]:
        """
        Override this function to provide the wallet a non-Polaris endpoint
        to begin the interactive flow. If the `amount` or `callback` arguments
        are not ``None``, make sure you include them in the URL returned.

        Note that ``after_interactive_flow()`` should also be implemented if this
        function is implemented to ensure the status, amounts, fees, and any
        other relevant information is saved to the ``Transaction`` record once
        the interactive flow is complete.

        :return: a URL to be used as the entry point for the interactive
            deposit flow
        """
        raise NotImplementedError()

    def after_interactive_flow(self, request: Request, transaction: Transaction):
        """
        Override this function to update the transaction or any related models with
        information collected during an external interactive flow.

        This function will be called each time the external application makes a
        request to the ``GET /sep24/transactions/deposit/interactive`` endpoint. This
        gives anchors the freedom to update the transaction once at the end of the flow
        or multiple times throughout the flow.

        The last time this function is called, the transaction's ``status`` value must
        be updated to ``pending_user_transfer_start`` or ``pending_anchor``, and the
        transaction's amounts and fees must also be updated. This will signal to the
        wallet application that the interactive flow has complete and that the anchor is
        ready to proceed.
        """
        raise NotImplementedError()

    def save_sep9_fields(
        self,
        token: SEP10Token,
        request: Request,
        stellar_account: str,
        fields: Dict,
        language_code: str,
        muxed_account: Optional[str] = None,
        account_memo: Optional[str] = None,
        account_memo_type: Optional[str] = None,
        *args: List,
        **kwargs: Dict
    ):
        """
        **DEPRECATED:** `stellar_account`, `account_memo`, `account_memo_type`, and `muxed_account`
        parameters. Use the `token` object passed instead.

        Save the `fields` passed for the user identified by `stellar_account` to pre-populate
        the forms returned from ``form_for_transaction()``. Note that this function is called
        before the transaction is created.

        For example, you could save the user's contact information with the model used
        for KYC information.
        ::

            # Assuming you have a similar method and model
            if token.muxed_account:
                user_key = token.muxed_account
            elif token.memo:
                user_key = f"{token.account}:{token.memo}"
            else:
                user_key = token.account
            user = user_for_key(user_key)
            user.phone_number = fields.get('mobile_number')
            user.email = fields.get('email_address')
            user.save()

        Then when returning a form to collect KYC information, also return the values
        saved in this method relevant to that form.
        ::

            # In your form_for_transaction() implementation
            if token.muxed_account:
                user_key = token.muxed_account
            elif token.memo:
                user_key = f"{token.account}:{token.memo}"
            else:
                user_key = token.account
            user = user_for_key(user_key)
            form_args = {
                'phone_number': format_number(user.phone_number),
                'email': user.email_address
            }
            return KYCForm(initial=form_args),

        If you'd like to validate the values passed in `fields`, you can perform any necessary
        checks and raise a ``ValueError`` in this function. Polaris will return the message of
        the exception in the response along with 400 HTTP status. The error message should be
        in the language specified by `language_code` if possible.
        """
        raise NotImplementedError()

    def process_sep6_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args: List,
        **kwargs: Dict
    ) -> Dict:
        """
        .. _deposit: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#deposit
        .. _deposit-exchange: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#deposit-exchange
        .. _Deposit no additional information needed: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#1-success-no-additional-information-needed
        .. _Customer information needed: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#2-customer-information-needed-non-interactive
        .. _Customer Information Status: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#4-customer-information-status

        This function is called during requests made to the SEP-6 deposit_ and
        `deposit-exchange`_ endpoints. The `params` object will contain the parameters
        included in the request. Note that Polaris will only call this function for
        `deposit-exchange`_ requests if SEP-38 is added to Polaris' ``ACTIVE_SEPS`` setting
        and the requested Stellar asset is enabled for SEP-38.

        If a request to the the `deposit-exchange`_ endpoint is made, a ``Quote`` object
        will be assigned to the ``Transaction`` object passed.

        Process these parameters and return one of the following responses outlined below
        as a dictionary. Save `transaction` to the DB if you plan to return a success
        response. If `transaction` is saved to the DB but a failure response is returned,
        Polaris will return a 500 error to the user.

        If you'd like the user to send ``Transaction.amount_in`` `plus the fee amount`,
        add the amount charged as a fee to ``Transaction.amount_in`` and
        ``Transaction.amount_expected``. here. While not required per SEP-6, it is
        encouraged to also populate ``Transaction.amount_fee`` and ``Transaction.amount_out``
        here as well. If this function is called for a `deposit-exchange`_ request,
        ``Transaction.fee_asset`` should also be assigned. If not assigned here, these
        columns must be assigned before returning the transaction from
        ``RailsIntegration.poll_pending_deposits()``.

        Note that the amount sent over the Stellar Network could differ from
        the amount specified in this API call, so fees and the amount delievered may have to
        be recalculated in ``RailsIntegration.poll_pending_deposits()`` for deposits and
        ``RailsIntegration.execute_outgoing_transaction()`` for withdrawals.

        Polaris responds to requests with the standard status code according the SEP. However,
        if you would like to return a custom error code in the range of 400-599 you may raise
        an ``rest_framework.exceptions.APIException``. For example, you could return a 503
        status code by raising an ``APIException("service unavailable", status_code=503)``.

        `Deposit no additional information needed`_

        The success response. Polaris creates most of the attributes described in this response.
        Simply return the 'how' and optionally 'extra_info' attributes. For example:
        ::

            return {
                "how": "<your bank account address>",
                "extra_info": {
                    "message": "Deposit the funds to the bank account specified in 'how'"
                }
            }

        `Customer information needed`_

        A failure response. Return the response as described in SEP.
        ::

            return {
              "type": "non_interactive_customer_info_needed",
              "fields" : ["family_name", "given_name", "address", "tax_id"]
            }

        `Customer Information Status`_

        A failure response. Return the 'type' and 'status' attributes. If
        ``CustomerIntegration.more_info_url()`` is implemented, Polaris will include the
        'more_info_url' attribute in the response as well.
        ::

            return {
              "type": "customer_info_status",
              "status": "denied",
            }

        :param token: the ``SEP10Token`` object representing the authenticated session
        :param request: a ``rest_framwork.request.Request`` object
        :param params: the request parameters as described in /deposit_
        :param transaction: an unsaved ``Transaction`` object representing the transaction
            to be processed
        """
        raise NotImplementedError(
            "`process_sep6_request` must be implemented if SEP-6 is active"
        )

    def create_channel_account(
        self, transaction: Transaction, *args: List, **kwargs: Dict
    ):
        """
        Create a temporary, per-deposit-transaction-object Stellar account using a different
        Stellar account and save the secret key of the created account to ``transaction.channel_seed``.

        This channel account must only be used as the source account for transactions related to the
        ``Transaction`` object passed. It also must not be used to submit transactions by any service
        other than Polaris. If it is, the outstanding transactions will be invalidated due to bad
        sequence numbers.

        If this integration function is called, the deposit payment represented by ``transaction``
        requires multiple signatures in order to be successfully submitted to the Stellar network.
        The anchored asset's distribution account may or may not be in that set of signatures
        required, depending on the configuration of the distribution account's signers.

        Once the transaction's signatures have been collected and the updated XDR written to
        ``transaction.envelope_xdr``, ``transaction.pending_signatures`` should be updated to
        ``False``, which will cause the ``process_pending_deposits`` process to submit it to the
        network along with the other transactions deemed ready by the anchor.

        If ``transaction.to_address`` doesn't exist on Stellar and the transaction has a
        ``channel_account``, ``transaction.channel_account`` will also be used to create and
        fund the destination account for the deposit payment to the user. So the channel
        account will be used for one or `potentially` two Stellar transactions.

        Once the deposit payment has been made on Stellar, Polaris will call ``after_deposit()``,
        which is where the anchor should merge the funds within ``transaction.channel_account``
        back to a persistent Stellar account owned by the anchor. See ``after_deposit()`` for
        more information.

        :param transaction: An object representing the transaction that requires a channel
            account as it's source.
        """
        raise NotImplementedError()

    def patch_transaction(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args: List,
        **kwargs: Dict
    ):
        """
        .. _`GET /info response`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#response-2
        .. _`GET /deposit`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#deposit
        .. _`GET /transactions`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#transaction-history
        .. _`PATCH /transactions`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#update

        Currently only used for SEP-6 transactions.

        The `GET /info response`_ contains a `fields` object that describes the custom fields
        an anchor requires in requests to the `GET /deposit`_ endpoint. If one or more of these fields
        were originally accepted but later discovered to be invalid in some way, an anchor can place
        the transaction in the ``pending_transaction_info_update`` status, save a JSON-serialized
        object describing the fields that need updating to ``Transaction.required_info_updates``,
        and save a human-readable message to ``Transaction.required_info_message`` describing the
        reason the fields need to be updated. A client can make a request to `GET /transactions`_ to
        detect that updated information is needed.

        This function is called when SEP-6 `PATCH /transactions`_ requests are made by the client to
        provide the updated values described in the ``Transaction.required_info_updates``. Use the
        `params` passed in the request to update `transaction` or any related data.

        Polaris validates that every field listed in ``Transaction.required_info_updates`` is
        present in `params` but cannot validate the values. If a ``ValueError`` is raised, Polaris
        will return a 400 response containing the exception message in the body.

        If no exception is raised, Polaris assumes the update was successful and will update the
        transaction's status back to ``pending_anchor`` as well as clear the ``required_info_updates``
        and ``required_info_message`` fields.

        :param token: the ``SEP10Token`` object representing the authenticated session
        :param request: a ``rest_framwork.request.Request`` object
        :param params: The request parameters as described in the `PATCH /transactions`_ endpoint.
        :param transaction: A ``Transaction`` object for which updated was provided.
        """
        raise NotImplementedError("PATCH /transactions/:id is not supported")


class WithdrawalIntegration:
    """
    The container class for withdrawal integration functions

    Subclasses must be registered with Polaris by passing it to
    ``polaris.integrations.register_integrations``.
    """

    def form_for_transaction(
        self,
        request: Request,
        transaction: Transaction,
        post_data: Optional[QueryDict] = None,
        amount: Optional[Decimal] = None,
        *args: List,
        **kwargs: Dict
    ) -> Optional[forms.Form]:
        """
        .. _SEP-9: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0009.md

        Same as ``DepositIntegration.form_for_transaction``

        :param request: a ``rest_framwork.request.Request`` object
        :param transaction: the ``Transaction`` database object
        :param post_data: the data included in the POST request body as a dictionary
        :param amount: a ``Decimal`` object the wallet may pass in the GET request.
            Use it to pre-populate your TransactionForm along with any SEP-9_
            parameters.
        """
        if transaction.amount_in:
            # we've collected transaction info
            # and don't implement KYC by default
            return
        elif post_data:
            return TransactionForm(transaction, post_data)
        else:
            return TransactionForm(transaction, initial={"amount": amount})

    def content_for_template(
        self,
        request: Request,
        template: Template,
        form: Optional[forms.Form] = None,
        transaction: Optional[Transaction] = None,
        *args: List,
        **kwargs: Dict
    ) -> Optional[Dict]:
        """
        Same as ``DepositIntegration.content_for_template``.

        :param request: a ``rest_framework.request.Request`` instance
        :param template: a ``polaris.templates.Template`` enum value
            for the template to be rendered in the response
        :param form: the form to be rendered in the template
        :param transaction: the transaction being processed
        """
        raise NotImplementedError()

    def after_form_validation(
        self,
        request: Request,
        form: forms.Form,
        transaction: Transaction,
        *args: List,
        **kwargs: Dict
    ):
        """
        Same as ``DepositIntegration.after_form_validation``, except
        `transaction.to_address` should be saved here when present in `form`.

        :param request: a ``rest_framework.request.Request`` instance
        :param form: the completed ``forms.Form`` submitted by the user
        :param transaction: the ``Transaction`` database object
        """
        raise NotImplementedError()

    def interactive_url(
        self,
        request: Request,
        transaction: Transaction,
        asset: Asset,
        amount: Optional[Decimal],
        callback: Optional[str],
        lang: Optional[str],
        *args: List,
        **kwargs: Dict
    ) -> Optional[str]:
        """
        Same as ``DepositIntegration.interactive_url``

        :return: a URL to be used as the entry point for the interactive
            withdrawal flow
        """
        raise NotImplementedError()

    def after_interactive_flow(self, request: Request, transaction: Transaction):
        """
        Same as ``DepositIntegration.after_interactive_flow``
        """
        raise NotImplementedError()

    def save_sep9_fields(
        self,
        token: SEP10Token,
        request: Request,
        stellar_account: str,
        fields: Dict,
        language_code: str,
        muxed_account: Optional[str] = None,
        account_memo: Optional[str] = None,
        account_memo_type: Optional[str] = None,
        *args: List,
        **kwargs: Dict
    ):
        """
        Same as ``DepositIntegration.save_sep9_fields``
        """
        raise NotImplementedError()

    def process_sep6_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args: List,
        **kwargs: Dict
    ) -> Dict:
        """
        .. _/withdraw: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#withdraw
        .. _Withdraw no additional information needed: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#1-success-no-additional-information-needed-1

        Same as ``DepositIntegration.process_sep6_request`` except for the case below.
        Specifically, the ``how`` attribute should not be included.

        `Withdraw no additional information needed`_

        A success response. Polaris populates most of the attributes for this response.
        Simply return an 'extra_info' attribute if applicable:
        ::

            {
                "extra_info": {
                    "message": "Send the funds to the following stellar account including 'memo'"
                }
            }

        In addition to this response, you may also return the `Customer information needed`_
        and `Customer Information Status`_ responses as described in
        ``DepositIntegration.process_sep6_request``.
        """
        raise NotImplementedError(
            "`process_sep6_request` must be implemented if SEP-6 is active"
        )

    def patch_transaction(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args: List,
        **kwargs: Dict
    ):
        """
        Same as ``DepositIntegration.patch_transaction``
        """
        raise NotImplementedError("PATCH /transactions/:id is not supported")


registered_deposit_integration = DepositIntegration()
registered_withdrawal_integration = WithdrawalIntegration()



from django.urls import re_path
from polaris.sep31 import info, transactions

urlpatterns = [
    re_path(r"^info/?$", info.info),
    re_path(
        r"^transactions/(?P<transaction_id>[^/]+)/?$",
        transactions.TransactionsAPIView.as_view(),
    ),
    re_path(r"^transactions/?$", transactions.TransactionsAPIView.as_view()),
]



"""
This module implements the logic for the `/.well-known` endpoint.
In particular, this generates the `.toml` file for the anchor server.

The importance of the Stellar TOML file is explained in SEP-0001:
https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0001.md
"""

import os

import toml
from django.contrib.staticfiles import finders
from django.utils.encoding import smart_str
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.renderers import BaseRenderer
from rest_framework.decorators import api_view, renderer_classes

from polaris import settings
from polaris.utils import getLogger
from polaris.integrations import (
    registered_toml_func,
    get_stellar_toml,
    registered_custody_integration as rci,
)
from polaris.models import Asset


logger = getLogger(__name__)


class PolarisPlainTextRenderer(BaseRenderer):
    """
    .. _documentation: https://www.django-rest-framework.org/api-guide/renderers/#custom-renderers

    Source copied from django-rest-framework's documentation_
    """

    media_type = "text/plain"
    format = "txt"

    def render(self, data, media_type=None, renderer_context=None):
        return smart_str(data, encoding=self.charset)


@api_view(["GET"])
@renderer_classes([PolarisPlainTextRenderer])
def generate_toml(request: Request) -> Response:
    """Generate a TOML-formatted string"""
    if registered_toml_func is get_stellar_toml:
        # integration function is not used, check to see if a static file is defined
        static_toml = None
        if settings.LOCAL_MODE:
            static_toml = finders.find("polaris/local-stellar.toml")
        if not static_toml:
            static_toml = finders.find("polaris/stellar.toml")
        if static_toml:
            with open(static_toml) as f:
                return Response(f.read(), content_type="text/plain")

    # The anchor uses the registered TOML function, replaced or not
    toml_dict = {
        "NETWORK_PASSPHRASE": settings.STELLAR_NETWORK_PASSPHRASE,
    }
    distribution_accounts = []
    for asset in Asset.objects.all():
        try:
            distribution_accounts.append(rci.get_distribution_account(asset))
        except NotImplementedError:
            break
    if distribution_accounts:
        toml_dict["ACCOUNTS"] = distribution_accounts
    if "sep-24" in settings.ACTIVE_SEPS:
        toml_dict["TRANSFER_SERVER"] = os.path.join(settings.HOST_URL, "sep24")
        toml_dict["TRANSFER_SERVER_SEP0024"] = toml_dict["TRANSFER_SERVER"]
    if "sep-6" in settings.ACTIVE_SEPS:
        toml_dict["TRANSFER_SERVER"] = os.path.join(settings.HOST_URL, "sep6")
    if "sep-10" in settings.ACTIVE_SEPS:
        toml_dict["WEB_AUTH_ENDPOINT"] = os.path.join(settings.HOST_URL, "auth")
        toml_dict["SIGNING_KEY"] = settings.SIGNING_KEY
    if "sep-12" in settings.ACTIVE_SEPS:
        toml_dict["KYC_SERVER"] = os.path.join(settings.HOST_URL, "kyc")
    if "sep-31" in settings.ACTIVE_SEPS:
        toml_dict["DIRECT_PAYMENT_SERVER"] = os.path.join(settings.HOST_URL, "sep31")
    if "sep-38" in settings.ACTIVE_SEPS:
        toml_dict["ANCHOR_QUOTE_SERVER"] = os.path.join(settings.HOST_URL, "sep38")

    toml_dict.update(registered_toml_func(request))
    content = toml.dumps(toml_dict)

    return Response(content, content_type="text/plain")



from django.db import migrations, models
import polaris.models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0002_auto_20200921_2135"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="asset", name="distribution_account_master_signer"
        ),
        migrations.RemoveField(model_name="asset", name="distribution_account_signers"),
        migrations.RemoveField(
            model_name="asset", name="distribution_account_thresholds"
        ),
    ]



from typing import Dict, Optional, List

from rest_framework.request import Request

from polaris.models import Asset


def default_info_func(
    request: Request,
    asset: Asset,
    lang: Optional[str],
    exchange: bool,
    *args: List,
    **kwargs: Dict
) -> Dict:
    """
    .. _/info response: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#response-2

    Replace this function with another by passing it to ``register_integrations()``
    as described in :ref:`api:Register Integrations`.

    Return a dictionary containing the `fields` and `types` key-value pairs
    described in the SEP-6 /info response for the asset passed. Raise a
    ``ValueError()`` if `lang` is not supported. For example,
    ::

        if asset.code == "USD":
            return {
                "fields": {
                    "email_address" : {
                        "description": "your email address for transaction status updates",
                        "optional": True
                    },
                    "amount" : {
                        "description": "amount in USD that you plan to deposit"
                    },
                    "type" : {
                        "description": "type of deposit to make",
                        "choices": ["SEPA", "SWIFT", "cash"]
                    }
                },
                "types": {
                    "bank_account": {
                        "fields": {
                            "dest": {"description": "your bank account number" },
                            "dest_extra": { "description": "your routing number" },
                            "bank_branch": { "description": "address of your bank branch" },
                            "phone_number": { "description": "your phone number in case there's an issue" }
                        }
                    },
                    "cash": {
                        "fields": {
                            "dest": {
                                "description": "your email address. Your cashout PIN will be sent here.",
                                "optional": True
                            }
                        }
                    }
                }
            }

    :param request: a ``rest_framework.request.Request`` object
    :param asset: ``Asset`` object for which to return the `fields` and `types`
        key-value pairs
    :param lang: the language code the client requested for the `description`
        values in the response
    :param exchange: whether or not the info returned will be used for a
        `deposit-exchange` or `withdraw-exchange` response object. Only
        relevant if SEP-38 is enabled.
    """
    raise NotImplementedError()


registered_info_func = default_info_func



import sys
import signal
import time
from decimal import Decimal
from datetime import datetime, timezone

import django.db.transaction
from django.db.models import Q
from django.core.management import BaseCommand

from polaris import settings
from polaris.integrations import registered_fee_func, calculate_fee
from polaris.utils import getLogger, maybe_make_callback
from polaris.models import Transaction
from polaris.integrations import registered_rails_integration as rri


logger = getLogger(__name__)
DEFAULT_INTERVAL = 30
TERMINATE = False


class Command(BaseCommand):
    """
    This process periodically queries for transactions that are ready to be executed
    off-chain and calls Polaris’
    :meth:`~polaris.integrations.RailsIntegration.execute_outgoing_transaction`
    integration function for each one. Ready transactions are those in
    ``pending_receiver`` or ``pending_anchor`` statuses, among other conditions.

    Anchors are expected to update the :attr:`~polaris.models.Transaction.status` to
    ``completed`` or ``pending_external`` if initiating the transfer was successful.

    **Optional arguments:**

        -h, --help            show this help message and exit
        --loop                Continually restart command after a specified number
                              of seconds.
        --interval INTERVAL, -i INTERVAL
                              The number of seconds to wait before restarting
                              command. Defaults to 30.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    @staticmethod
    def exit_gracefully(sig, frame):  # pragma: no cover
        logger.info("Exiting execute_outgoing_transactions...")
        module = sys.modules[__name__]
        module.TERMINATE = True

    @staticmethod
    def sleep(seconds):  # pragma: no cover
        module = sys.modules[__name__]
        for _ in range(seconds):
            if module.TERMINATE:
                break
            time.sleep(1)

    def add_arguments(self, parser):  # pragma: no cover
        parser.add_argument(
            "--loop",
            action="store_true",
            help="Continually restart command after a specified number of seconds.",
        )
        parser.add_argument(
            "--interval",
            "-i",
            type=int,
            help=(
                "The number of seconds to wait before restarting command. "
                "Defaults to {}.".format(DEFAULT_INTERVAL)
            ),
        )

    def handle(self, *_args, **options):  # pragma: no cover
        module = sys.modules[__name__]
        if options.get("loop"):
            while True:
                if module.TERMINATE:
                    break
                self.execute_outgoing_transactions()
                self.sleep(options.get("interval") or DEFAULT_INTERVAL)
        else:
            self.execute_outgoing_transactions()

    @staticmethod
    def execute_outgoing_transactions():
        """
        Execute pending withdrawals.
        """
        module = sys.modules[__name__]
        sep31_qparams = Q(
            protocol=Transaction.PROTOCOL.sep31,
            status=Transaction.STATUS.pending_receiver,
            kind=Transaction.KIND.send,
        )
        sep6_24_qparams = Q(
            protocol__in=[Transaction.PROTOCOL.sep24, Transaction.PROTOCOL.sep6],
            status=Transaction.STATUS.pending_anchor,
            kind__in=[
                Transaction.KIND.withdrawal,
                getattr(Transaction.KIND, "withdrawal-exchange"),
            ],
        )
        with django.db.transaction.atomic():
            transactions = list(
                Transaction.objects.filter(
                    sep6_24_qparams | sep31_qparams, pending_execution_attempt=False
                ).select_for_update()
            )
            ids = []
            for t in transactions:
                t.pending_execution_attempt = True
                ids.append(t.id)
            Transaction.objects.filter(id__in=ids).update(
                pending_execution_attempt=True
            )

        if transactions:
            logger.info(f"Executing {len(transactions)} outgoing transactions")
        num_completed = 0
        for i, transaction in enumerate(transactions):
            if module.TERMINATE:
                still_processing_transactions = transactions[i:]
                Transaction.objects.filter(
                    id__in=[t.id for t in still_processing_transactions]
                ).update(pending_execution_attempt=False)
                break

            logger.info(f"Calling execute_outgoing_transaction() for {transaction.id}")
            try:
                rri.execute_outgoing_transaction(transaction)
            except NotImplementedError:
                logger.error(
                    "RailsIntegration.execute_outgoing_transaction() is not implemented"
                )
                module.TERMINATE = True
                break
            except Exception:
                transaction.pending_execution_attempt = False
                transaction.save()
                logger.exception(
                    "execute_outgoing_transactions() threw an unexpected exception"
                )
                continue

            transaction.refresh_from_db()
            if (
                transaction.protocol == Transaction.PROTOCOL.sep31
                and transaction.status == Transaction.STATUS.pending_receiver
            ) or (
                transaction.protocol
                in [Transaction.PROTOCOL.sep24, Transaction.PROTOCOL.sep6]
                and transaction.status == transaction.STATUS.pending_anchor
            ):
                transaction.pending_execution_attempt = False
                if transaction.quote:
                    transaction.quote.save()
                transaction.save()
                logger.error(
                    f"Transaction {transaction.id} status must be "
                    f"updated after call to execute_outgoing_transaction()"
                )
                continue
            elif transaction.status in [
                Transaction.STATUS.pending_external,
                Transaction.STATUS.completed,
            ]:
                if transaction.amount_fee is None or transaction.amount_out is None:
                    if transaction.quote:
                        err_msg = (
                            f"transaction {transaction.id} uses a quote but was returned "
                            "from execute_outgoing_transaction() without amount_fee or amount_out "
                            "assigned, skipping"
                        )
                        logger.error(err_msg)
                        transaction.message = err_msg
                        transaction.pending_execution_attempt = False
                        transaction.quote.save()
                        transaction.save()
                        continue
                    logger.warning(
                        f"transaction {transaction.id} was returned from execute_outgoing_transaction() "
                        "without Transaction.amount_fee or Transaction.amount_out assigned. Future Polaris "
                        "releases will not calculate fees and delivered amounts."
                    )
                if transaction.amount_fee is None:
                    if not transaction.quote and registered_fee_func is calculate_fee:
                        op = {
                            Transaction.KIND.withdrawal: settings.OPERATION_WITHDRAWAL,
                            getattr(
                                Transaction.KIND, "withdrawal-exchange"
                            ): settings.OPERATION_WITHDRAWAL,
                            Transaction.KIND.send: settings.OPERATION_SEND,
                        }[transaction.kind]
                        try:
                            transaction.amount_fee = calculate_fee(
                                {
                                    "amount": transaction.amount_in,
                                    "operation": op,
                                    "asset_code": transaction.asset.code,
                                }
                            )
                        except ValueError:
                            transaction.pending_execution_attempt = False
                            transaction.save()
                            logger.exception("Unable to calculate fee")
                            continue
                    else:
                        transaction.amount_fee = Decimal(0)
                if not transaction.quote:
                    transaction.amount_out = round(
                        transaction.amount_in - transaction.amount_fee,
                        transaction.asset.significant_decimals,
                    )
                # Anchors can mark transactions as pending_external if the transfer
                # cannot be completed immediately due to external processing.
                # poll_outgoing_transactions will check on these transfers and mark them
                # as complete when the funds have been received by the user.
                if transaction.status == Transaction.STATUS.completed:
                    num_completed += 1
                    transaction.completed_at = datetime.now(timezone.utc)
            elif transaction.status not in [
                Transaction.STATUS.error,
                Transaction.STATUS.pending_transaction_info_update,
                Transaction.STATUS.pending_customer_info_update,
            ]:
                transaction.pending_execution_attempt = False
                if transaction.quote:
                    transaction.save()
                transaction.save()
                logger.error(
                    f"Transaction {transaction.id} was moved to invalid status"
                    f" {transaction.status}"
                )
                continue

            transaction.pending_execution_attempt = False
            if transaction.quote:
                transaction.quote.save()
            transaction.save()
            maybe_make_callback(transaction)

        if num_completed:
            logger.info(f"{num_completed} transfers have been completed")



import os
from django.apps import AppConfig
from django.core.exceptions import ImproperlyConfigured


class PolarisConfig(AppConfig):
    name = "polaris"
    verbose_name = "Django Polaris"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        """
        Initialize the app
        """
        from decimal import setcontext, DefaultContext
        from polaris import settings  # loads internal settings
        from polaris import cors  # loads CORS signals
        from polaris.sep24.utils import check_sep24_config

        # Set in-memory precision to match database-level precision
        # https://adamj.eu/tech/2020/03/23/setting-pythons-decimal-context-for-all-threads/#in-django
        DefaultContext.prec = 30
        setcontext(DefaultContext)

        self.check_middleware()
        self.check_protocol()
        if "sep-24" in settings.ACTIVE_SEPS:
            check_sep24_config()

    @staticmethod
    def check_middleware():
        from django.conf import settings as django_settings

        cors_middleware_path = "corsheaders.middleware.CorsMiddleware"
        if cors_middleware_path not in django_settings.MIDDLEWARE:
            raise ImproperlyConfigured(
                f"{cors_middleware_path} is not installed in settings.MIDDLEWARE"
            )

    @staticmethod
    def check_protocol():
        from polaris import settings
        from polaris.utils import getLogger
        from django.conf import settings as django_settings

        logger = getLogger(__name__)
        if getattr(django_settings, "SECURE_PROXY_SSL_HEADER"):
            logger.debug(
                "SECURE_PROXY_SSL_HEADER should only be set if Polaris is "
                "running behind an HTTPS reverse proxy."
            )
        elif not (
            settings.LOCAL_MODE or getattr(django_settings, "SECURE_SSL_REDIRECT")
        ):
            logger.debug(
                "SECURE_SSL_REDIRECT is required to redirect HTTP traffic to HTTPS"
            )



"""This module defines the URL patterns for the `/.well-known/stellar.toml` endpoint."""
from django.urls import re_path

from polaris.sep1.views import generate_toml

urlpatterns = [re_path(r"^stellar.toml/?$", generate_toml)]
